LOGLEVELS = {'debug': 10,
             'info': 20,
             'warning': 30,
             'error': 40,
             'critical': 50}

def _init_logger():
    import logging
    ### LOGGING
    # Note: configuration will be propagated to all child modules of
    # pyaerocom, for details see
    # http://eric.themoritzfamily.com/learning-python-logging.html
    logger = logging.getLogger('pyaerocom')

    default_formatter = logging.Formatter(\
       "%(asctime)s:%(levelname)s:\n%(message)s")

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(default_formatter)

    logger.addHandler(console_handler)

    logger.setLevel(logging.CRITICAL)

    print_log = logging.getLogger('pyaerocom_print')

    print_handler = logging.StreamHandler()
    print_handler.setFormatter(logging.Formatter("%(message)s"))

    print_log.addHandler(print_handler)

    print_log.setLevel(logging.INFO)
    return (logger, print_log)

def change_verbosity(new_level='debug', log=None):
    """
    Change verbosity of one of the pyaerocom loggers

    Parameters
    ----------
    new_level : str or int
        choose from either keys, or values of :attr:`LOGLEVELS`.
    log
        either `pyaerocom.logger` or `pyaerocom.print_log`.

    Returns
    -------
    None

    """
    if log is None:
        from pyaerocom import logger
        log = logger
    if isinstance(new_level, str):
        if not new_level in LOGLEVELS:
            raise ValueError(f'invalid log level {new_level}, choose from '
                             f'keys or values of {LOGLEVELS}')
        new_level = LOGLEVELS[new_level]
    else:
        if not new_level in LOGLEVELS.values():
            raise ValueError(f'invalid log level {new_level}, choose from '
                             f'keys or values of {LOGLEVELS}')
    log.setLevel(new_level)

### Functions for package initialisation
def _init_supplemental():
    """
    Get version and pyaerocom installation path

    Returns
    -------
    str
        version string
    str
        path to source code base directory (
        <installed_basedir>/pyaerocom/pyaerocom)


    """
    from pkg_resources import get_distribution
    import os
    dist = get_distribution('pyaerocom')
    return (dist.version, os.path.join(dist.location, 'pyaerocom'))





"""
Pyearocom module for reading and processing of EBAS NASA Ames files

For details on the file format see `here <https://ebas-submit.nilu.no/
Submit-Data/Getting-started>`__
"""
import os
import numpy as np
import pandas as pd
from collections import OrderedDict as od
from datetime import datetime
from pyaerocom._lowlevel_helpers import str_underline, dict_to_str
from pyaerocom.exceptions import (TimeZoneError, NasaAmesReadError)
from pyaerocom import const

class EbasColDef(dict):
    """Dict-like object for EBAS NASA Ames column definitions

    Note
    ----
    The meta attribute name 'unit' can also be accessed using the CF attr name
    'units'

    Attributes
    ----------
    name : str
        column name
    unit : str
        unit of data in column (if applicable)
    is_var : bool
        True if column corresponds to variable data, False if not
    is_flag : bool
        True, if column corresponds to Flag column, False if not
    flag_col : int
        column number of flag column that corresponds to this data column (only
        relevant if :attr:`is_var` is True)

    Parameters
    ----------
    name : str
        column name
    is_var : bool
        True if column corresponds to variable data, False if not
    is_flag : bool
        True, if column corresponds to Flag column, False if not
    unit : :obj:`str`, optional
        unit of data in column (if applicable)
    flag_col : :obj:`str`, optional
        ``name`` of flag column that corresponds to this data colum (only
        relevant if :attr:`is_var` is True)
    """
    def __init__(self, name, is_var, is_flag, unit="1"):
        self.name = name
        self.unit = unit
        self.is_var = is_var
        self.is_flag = is_flag
        self.flag_col = None

    def get_wavelength_nm(self):
        """Try to access wavelength information in nm (as float)"""
        if not 'wavelength' in self:
            raise KeyError('Column variable {} does not contain wavelength '
                           'information'.format(self.name))
        elif not 'nm' in self.wavelength:
            raise NotImplementedError('Wavelength definition is not in nm')
        return float(self.wavelength.split('nm')[0].strip())

    def to_dict(self, ignore_keys=['is_var', 'is_flag', 'flag_col',
                                   'wavelength_nm']):
        d = {}
        for k, v in self.items():
            if k in ignore_keys:
                continue
            elif k == 'unit':
                k ='units'
            d[k] = v
        return d

    def __getitem__(self, key): # add support for units attr (CF standard name)
        if key == 'units':
            key = 'unit'
        return super(EbasColDef, self).__getitem__(key)

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, val):
        self[key] = val

    def __repr__(self):
        return str(self)

    def __str__(self):
        colattrs = ''
        for key, val in self.items():
            colattrs += f'{key}: {val}, '
        return colattrs[:-2]

def _readline_ref_and_revision(line):
    spl = line.strip().split()
    basedate = np.datetime64(datetime.strptime('{}{}{}'.format(spl[0],
                                               spl[1], spl[2]), "%Y%m%d"), 's')
    rev = np.datetime64(datetime.strptime('{}{}{}'.format(spl[3],
                                               spl[4], spl[5]), "%Y%m%d"), 's')
    return [basedate, rev]

class NasaAmesHeader(object):
    """Header class for Ebas NASA Ames file

    Note
    ----
    Is used in :class:`EbasNasaAmesFile` and should not be used directly.

    Attributes
    -----------

    """
    _NUM_FIXLINES = 13
    _HEAD_ROWS_MANDATORY = [0,5,8,9,10,11]

    #conversion methods for first 13 header lines of
    CONV_STR = lambda l : str(l.strip())
    CONV_PI = lambda l : '; '.join([x.strip() for x in l.split(';')])
    CONV_MULTIINT = lambda l : [int(x) for x in l.strip().split()]
    CONV_MULTIFLOAT = lambda l : [float(x) for x in l.strip().split()]
    CONV_INT = lambda l : int(l.strip())
    CONV_FLOAT = lambda l : float(l.strip())
    _STARTDATE_FMT = "%Y%m%d%H%M%S"

    # the parameters that are always in the beginning of the file
    _H_FIXLINES_YIELD = [["num_head_lines", "num_head_fmt"], #line 1
                         "data_originator", #line 2
                         "sponsor_organisation", #3
                         "submitter", #4
                         "project_association", #5
                         ["vol_num", "vol_totnum"], #6
                         ["ref_date", "revision_date"], #7
                         "freq", #8
                         "descr_time_unit", #9
                         "num_cols_dependent", #10
                         "mul_factors", #11
                         "vals_invalid", #12
                         "descr_first_col", #13
                         ]
    # conversion methods are defined above
    _H_FIXLINES_CONV = [CONV_MULTIINT, #1 -> yields 2
                        CONV_PI, #2
                        CONV_STR, #3
                        CONV_STR, #4
                        CONV_STR, #5
                        CONV_MULTIINT, #6
                        _readline_ref_and_revision, #7 lambda l : [x.strip() for x in l.strip().split("     ")], #7
                        CONV_FLOAT, #8
                        CONV_STR, #9
                        CONV_INT, #10
                        CONV_MULTIFLOAT, #11
                        CONV_MULTIFLOAT, #12
                        CONV_STR] #13

    def __init__(self, **kwargs):
        self._head_fix = od(num_head_lines = np.nan,
                            num_head_fmt = np.nan,
                            data_originator = "",
                            sponsor_organisation = "",
                            submitter = "",
                            project_association = "",
                            vol_num = np.nan,
                            vol_totnum = np.nan,
                            ref_date = np.nan,
                            revision_date = np.nan,
                            freq = np.nan,
                            descr_time_unit = "",
                            num_cols_dependent = np.nan,
                            mul_factors = [],
                            vals_invalid = [],
                            descr_first_col = "")
        self._var_defs = []
        self._meta = od()
        self.update(**kwargs)

    @property
    def head_fix(self):
        """Dictionary containing fixed header info (that is always available)"""
        return self._head_fix

    @property
    def var_defs(self):
        """List containing column variable definitions

        List index is column index in file and value is instance of
        :class:`EbasColDef`
        """
        return self._var_defs

    @property
    def meta(self):
        """Meta data dictionary (specific for this file)"""
        return self._meta

    def update(self, **kwargs):
        for k, v in kwargs.items():
            self[k] = v

    def __getattr__(self, key):
        if key in self._head_fix:
            return self._head_fix[key]
        elif key in self._meta:
            return self._meta[key]
        else:
            raise AttributeError("Invalid attribute: {}".format(key))

    def __getitem__(self, key):
        return self.__getattr__(key)

    def __setitem__(self, key, val):
        if key in self._head_fix:
            self._head_fix[key] = val
        else:
            self._meta[key] = val

    def __str__(self):
        head = "{}".format(type(self).__name__)
        s = "{}\n{}\n".format(head, len(head)*"-")
        s += dict_to_str(self._head_fix)
        s += "\n\n{}".format(str_underline("Column variable definitions",
                                           indent=3))
        for item in self._var_defs:
            s += "\n   {}".format(repr(item))
        s += "\n\n{}".format(str_underline("EBAS meta data",
                                             indent=3))
        s += dict_to_str(self.meta)

        return s

class EbasFlagCol(object):
    """Simple helper class to decode and interpret EBAS flag columns

    Attributes
    ----------
    raw_data : ndarray
        raw flag column (containing X-digit floating point numbers)

    """
    def __init__(self, raw_data, interpret_on_init=True):
        self.raw_data = raw_data

        self._decoded = None
        self._valid = None

        if interpret_on_init:
            self.decode()

    @property
    def FLAG_INFO(self):
        """Detailed information about EBAS flag definitions"""
        return const.ebas_flag_info

    @property
    def decoded(self):
        """Nx3 numpy array containing decoded flag columns"""
        if self._decoded is None:
            self.decode()
        return self._decoded

    @property
    def valid(self):
        """Boolean array specifying valid and invalid measurements"""
        if self._valid is None:
            self.decode()
        return self._valid

    def decode(self):
        """Decode raw flag column"""
        flags = np.zeros((len(self.raw_data), 3)).astype(int)
        mask = self.raw_data.astype(bool) # rmooves all points that are 0, i.e. that contain no flag (valid measurements)
        valid = np.ones_like(self.raw_data).astype(bool)
        not_ok = self.raw_data[mask]
        if len(not_ok) > 0:
            _decoded = []
            _valid = []
            for flag in not_ok:
                item = "{:.9f}".format(flag).split(".")[1]
                vals = [int(item[:3]), int(item[3:6]), int(item[6:9])]
                _invalid = False
                for val in vals:
                    if val == 100:
                        _invalid = False
                        break # since all other flags are irrelevant if 100 is flagged
                    elif val in self.FLAG_INFO['valid'] and not self.FLAG_INFO['valid'][val]: #is invalid
                        _invalid = True

                _decoded.append(vals)
                _valid.append(not _invalid)

            flags[mask] = np.asarray(_decoded)
            valid[mask] = _valid

        self._valid = valid
        self._decoded = flags

class EbasNasaAmesFile(NasaAmesHeader):
    """EBAS NASA Ames file interface

    Class interface for reading and processing of EBAS NASA Ames file

    Attributes
    ----------
    time_stamps : ndarray
        array containing datetime64 objects with timestamps
    flags : dict
        dictionary containing :class:`EbasFlagCol` objects for each column
        containing flags

    Parameters
    ----------
    file : :obj:`str`, optional
        EBAS NASA Ames file. if valid file path, then the file is read on
        init (please note following options for import)
    only_head : bool
        read only file header
    replace_invalid_nan : bool
        replace all invalid values in the table by NaNs. The invalid values for
        each dependent data column are identified based on the information in
        the file header.
    convert_timestamps : bool
        compute array of numpy datetime64 timestamps from numeric timestamps
        in data
    evaluate_flags : bool
        if True, all flags in all flag columns are decoded from floating
        point representation to 3 integers, e.g.
        0.111222333 -> 111 222 333
    quality_check : bool
        perform quality check after import (for details see
        :func:`_quality_check`)
    **kwargs
        optional input args that are passed to init of :class:`NasaAmesHeader`
        base class

    """
    TIMEUNIT2SECFAC = dict(days = 3600*24,
                           Days = 3600*24)

    ERR_LOW_STATS = 'percentile:15.87'
    ERR_HIGH_STATS = 'percentile:84.13'
    def __init__(self, file=None, only_head=False, replace_invalid_nan=True,
                 convert_timestamps=True, evaluate_flags=False,
                 quality_check=True, **kwargs):
        super(EbasNasaAmesFile, self).__init__(**kwargs)
        self._data_header = [] #Header line of data block
        self._data = [] #data block

        self.time_stamps = None
        self.start_meas = None
        self.stop_meas = None

        self.flag_col_info = od()

        self.file = None

        if file is not None:
            if not os.path.exists(file):
                raise IOError("File {} does not exists".format(file))
            self.read_file(file, only_head, replace_invalid_nan,
                           convert_timestamps, evaluate_flags,
                           quality_check)

    @property
    def data(self):
        """2D numpy array containing data table"""
        return self._data

    @property
    def data_header(self):
        return self._data_header

    @property
    def shape(self):
        """Shape of data array"""
        return self.data.shape

    @property
    def col_num(self):
        """Number of columns in table"""
        return self.num_cols_dependent + 1

    @property
    def col_names(self):
        """Column names of table"""
        #cols = [x["name"] for x in self.var_defs]
        return [x["name"] for x in self.var_defs]

    @property
    def col_names_vars(self):
        """Names of all columns that are flagged as variables"""
        return [x.name for x in self.var_defs if x.is_var]

    @property
    def col_nums_vars(self):
        """Column index number of all variables"""
        return [i for (i, item) in enumerate(self.var_defs) if item.is_var]

    @property
    def base_date(self):
        """Base date of data as numpy.datetime64[s]"""
        if not "timezone" in self.meta:
            raise AttributeError("Fatal: could not infer base date. Timezone "
                                 "is not available in file header")
        if not self.timezone.lower() == "utc":
            raise TimeZoneError("Timezones other than UTC are not yet supported")
        return self.ref_date

    @property
    def time_unit(self):
        """Time unit of data"""
        return self.descr_time_unit.split()[0].strip()

    @staticmethod
    def numarr_to_datetime64(basedate, num_arr, mulfac_to_sec):
        """Convert array of numerical timestamps into datetime64 array

        Parameters
        ----------
        basedate : datetime64
            reference date
        num_arr : ndarray
            numerical time stamps relative to basedate
        mulfac_to_sec : float
            multiplicative factor to convert numerical values to unit of
            seconds
        Returns
        -------
        ndarray
            array containing timestamps as datetime64 objects
        """
        totnum = len(num_arr)
        if totnum == 0:
            raise AttributeError("No data available in file")
        return basedate + (num_arr * mulfac_to_sec).astype("timedelta64[s]")

    def all_cols_contain(self, colnums, what):
        """
        Check if all input columns contain input attr `what`

        Parameters
        ----------
        colnums : list
            list of column numbers
        what : str
            name of attribute (e.g. `matrix`, `statistics`,
            `tower_inlet_height`)

        Returns
        -------
        bool
            True if all input columns contain `what` attr., else False.

        """
        return all([what in self.var_defs[col] for col in colnums])

    def assign_flagcols(self):
        _prev = 0
        for (idx, item) in enumerate(self.var_defs):
            if item.is_flag:
                for _idx in range(_prev, idx):
                    self.var_defs[_idx].flag_col=idx
                _prev = idx + 1

    def init_flags(self, evaluate=True):
        """Decode flag columns and store info in :attr:`flags`
        """
        for (idx, item) in enumerate(self.var_defs):
            if item.is_flag:
                data = self.data[:, idx]
                flag = EbasFlagCol(raw_data=data,
                                   interpret_on_init=evaluate)
                self.flag_col_info[idx] = flag

    def compute_time_stamps(self):
        """Compute time stamps from first two data columns"""
        if not self.var_defs[0].unit == self.var_defs[1].unit:
            raise NasaAmesReadError('2nd column is not endtime or does not '
                                    'have the same unit as 1st column (starttime)')
        offs = self.base_date
        unit = self.time_unit
        if not unit in self.TIMEUNIT2SECFAC:
            raise ValueError("Invalid unit for temporal resolution: {}".format(unit))
        mulfac = self.TIMEUNIT2SECFAC[unit]

        start = self.numarr_to_datetime64(offs, self.data[:,0], mulfac)
        stop = self.numarr_to_datetime64(offs, self.data[:,1], mulfac)

        dt = stop - start
        # mid timestamps
        self.time_stamps = start + dt*.5
        self.start_meas = start
        self.stop_meas = stop

        return (start, stop)

    def get_time_gaps_meas(self, np_freq='s'):
        """Get array with time gaps between individual measurements

        This is computed based on start and stop timestamps, e.g.
        `=dt[0] = start[1] - stop[0]`

        Parameters
        ----------
        np_freq : str
            string specifying output frequency of gap values

        Returns
        -------
        ndarray
            array with time-differences as floating point number in specified
            input resolution
        """
        if self.start_meas is None:
            self.compute_time_stamps() # assigns start / stop attrs.
        start, stop = self.start_meas, self.stop_meas
        gaps = (start[1:] - stop[:-1]).astype(f'timedelta64[{np_freq}]').astype(float)
        return gaps

    def get_dt_meas(self, np_freq='s'):
        """Get array with time between individual measurements

        This is computed based on start and timestamps, e.g.
        `dt[0] = start[1] - start[0]`

        Parameters
        ----------
        np_freq : str
            string specifying output frequency of gap values

        Returns
        -------
        ndarray
            array with time-differences as floating point number in specified
            input resolution
        """
        if self.start_meas is None:
            self.compute_time_stamps() # assigns start / stop attrs.
        start = self.start_meas
        dts = (start[1:] - start[:-1]).astype(f'timedelta64[{np_freq}]').astype(float)
        return dts

    def _quality_check(self):
        msgs = ""
        if not len(self.data_header) == len(self.var_defs):
            msgs += ("Mismatch between variable definitions in header and "
                     "number of data columns in table\n")
        if not "timezone" in self.meta:
            msgs += ("Timezone not defined in metadata")
        if msgs:
            raise AttributeError("Quality check failed. Messages: {}".format(msgs))

    def read_header(self, nasa_ames_file, quality_check=True):
        self.read_file(nasa_ames_file, only_head=True,
                       quality_check=quality_check)

    def read_file(self, nasa_ames_file, only_head=False,
                  replace_invalid_nan=True, convert_timestamps=True,
                  evaluate_flags=False, quality_check=False):
        """Read NASA Ames file

        Parameters
        ----------
        nasa_ames_file : str
            EBAS NASA Ames file
        only_head : bool
            read only file header
        replace_invalid_nan : bool
            replace all invalid values in the table by NaNs. The invalid values for
            each dependent data column are identified based on the information in
            the file header.
        convert_timestamps : bool
            compute array of numpy datetime64 timestamps from numeric timestamps
            in data
        evaluate_flags : bool
            if True, all data columns get assigned their corresponding flag
            column, the flags in all flag columns are decoded from floating
            point representation to 3 integers, e.g. 0.111222333 -> 111 222 333
            and if input ```replace_invalid_nan==True```, then the invalid
            measurements in each column are replaced with NaN's.
        quality_check : bool
            perform quality check after import (for details see
            :func:`_quality_check`)
        """
        const.logger.info("Reading NASA Ames file:\n{}".format(nasa_ames_file))
        lc = 0 #line counter
        dc = 0 #data block line counter
        mc = 0 #meta block counter
        END_VAR_DEF = np.nan #will be set (info stored in header)
        IN_DATA = False
        data = []
        self.file = nasa_ames_file
        for line in open(nasa_ames_file):
            if IN_DATA: #in data block (end of file)
                try:
                    data.append(tuple([float(x.strip()) for x in line.strip().split()]))
                    #data.append([float(x.strip()) for x in line.strip().split()])
                except Exception as e:
                    const.print_log.warning(
                        f"EbasNasaAmesFile: Failed to read data row {dc}. "
                        f"Reason: {e}")
                dc += 1
            elif lc < self._NUM_FIXLINES: #in header section (before column definitions)
                try:
                    val = self._H_FIXLINES_CONV[lc](line)
                    attr = self._H_FIXLINES_YIELD[lc]
                    if isinstance(attr, list):
                        for i, attr_id in enumerate(attr):
                            self[attr_id] = val[i]
                    else:
                        self[attr] = val
                except Exception as e:
                    msg = ("Failed to read header row {}.\n{}\n"
                           "Error msg: {}".format(lc, line, repr(e)))
                    if lc in self._HEAD_ROWS_MANDATORY:
                        raise NasaAmesReadError("Fatal: {}".format(msg))
                    else:
                        const.logger.warning(msg)
            else: # behind header section and before data definition (contains column defs and meta info)
                if mc == 0: # still in column definition
                    END_VAR_DEF = self._NUM_FIXLINES + self.num_cols_dependent - 1
                    NUM_HEAD_LINES = self.num_head_lines
                    try:
                        self.var_defs.append(self._read_vardef_line(line))
                    except Exception as e:
                        const.logger.warning(repr(e))

                elif lc < END_VAR_DEF:
                    self.var_defs.append(self._read_vardef_line(line))

                elif lc == NUM_HEAD_LINES - 1:
                    IN_DATA = True
                    self._data_header = h = [x.strip() for x in line.split()]
                    #append information of first two columns to variable
                    #definition array.
                    self._var_defs.insert(0, EbasColDef(name=h[0],
                                                        is_flag=False,
                                                        is_var=False,
                                                        unit=self.time_unit))
                    self._var_defs.insert(1, EbasColDef(name=h[1],
                                                        is_flag=False,
                                                        is_var=False,
                                                        unit=self.time_unit))
                    if only_head:
                        return
                    const.logger.debug("REACHED DATA BLOCK")
                elif lc >= END_VAR_DEF + 2:
                    try:
                        name, val = line.split(":")
                        key = name.strip().lower().replace(" ", "_")
                        self.meta[key] = val.strip()
                    except Exception as e:
                        const.logger.warning("Failed to read line no. {}.\n{}\n"
                              "Error msg: {}\n".format(lc, line, repr(e)))
                else:
                    const.logger.debug("Ignoring line no. {}: {}".format(lc, line))
                mc += 1
            lc += 1

        data = np.asarray(data)

        data[:, 1:] = data[:, 1:] * np.asarray(self.mul_factors)

        self._data = data
        if replace_invalid_nan:
            dep_dat = data[:, 1:]
            for i, val in enumerate(np.floor(self.vals_invalid)):
                col = dep_dat[:, i]
                cond = np.floor(col) == val
                col[cond] = np.nan
                dep_dat[:, i] = col

            data[:, 1:] = dep_dat
        self._data = data

        if convert_timestamps:
            self.compute_time_stamps()

        self.assign_flagcols()
        self.init_flags(evaluate=evaluate_flags)

        if quality_check:
            self._quality_check()

    def _read_vardef_line(self, line_from_file):
        """Import variable definition line from NASA Ames file"""
        spl = [x.strip() for x in line_from_file.split(",")]
        name = spl[0]
        if not len(spl) > 1:
            unit = ''
        else:
            unit = spl[1]
        data = EbasColDef(name=name,
                          is_flag=True,
                          is_var=False,
                          unit=unit)

        if not "numflag" in name:
            data.is_var = True
            data.is_flag = False
            for item in spl[2:]:
                if "=" in item:
                    # e.g. wavelength=550nm
                    sub = item.split("=")
                    if len(sub) == 2:
                        idf, val = [x.strip() for x in sub]
                        data[idf.lower().replace(' ', '_')] = val
                    else:
                        const.logger.warning("Could not interpret part of column "
                                       "definition in EBAS NASA Ames file: "
                                       "{}".format(item))
                else: #unit
                    const.logger.warning("Failed to interpret {}".format(item))

        return data

    def _data_short_str(self):
        if len(self.data) == 0:
            s = "No data available"
        else:
            s =  str(self.data)
            shape = self.data.shape
            s += "\nColnum: {}".format(shape[1])
            s += "\nTimestamps: {}".format(shape[0])
        return s

    def print_col_info(self):
        """Print information about individual columns"""
        for (idx, coldef) in enumerate(self.var_defs):
            print("Column {}\n{}".format(idx, coldef))

    def __str__(self):
        s = super(EbasNasaAmesFile, self).__str__()
        s += "\n\n{}".format(str_underline("Data", indent=3))
        s += "\n{}".format(self._data_short_str())
        return s



import numpy as np
import os
import pandas as pd
#from datetime import datetime
from collections import OrderedDict

from pyaerocom import const
from pyaerocom.stationdata import StationData
from pyaerocom.ungriddeddata import UngriddedData
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
#from pyaerocom.io.helpers_units import (unitconv_sfc_conc, unitconv_wet_depo)

from pyaerocom.units_helpers import convert_unit
from pyaerocom.helpers import get_tot_number_of_seconds

class ReadAasEtal(ReadUngriddedBase):
    """Read subset of GAW-TAD-EANET data related to Aas et al., 2019

    Paper URL: https://www.nature.com/articles/s41598-018-37304-0

    See Also
    ---------
    :class:`ReadUngriddedBase`
	"""
    # name of files in GawTadSubsetAasEtAl
    _FILEMASK = '*.csv' # fix

    #: version log of this class (for caching)
    __version__ = '0.09'

    COL_DELIM = ','

    #: Temporal resoloution
    TS_TYPE = 'monthly'

    #: Name of dataset (OBS_ID)
    DATA_ID = 'GAWTADsubsetAasEtAl' #'GAWTADsubsetAasEtAl'

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [DATA_ID]

    #: value corresponding to invalid measurement
    #NAN_VAL = -9999.
    NAN_VAL = -999.
    #: List of variables that are provided by this dataset (will be extended
    #: by auxiliary variables on class init, for details see __init__ method of
    #: base class ReadUngriddedBase)
    # This contains the mapping between the requested variables and what it is called in the files.

    COLNAMES_VARS = {}
    COLNAMES_VARS['concentration_mgS/L']  = ['concso4pr']
    COLNAMES_VARS['concentration_ugS/m3'] = ['concso4', 'concso2']
    COLNAMES_VARS['precip_amount_mm']     = ['pr']
    COLNAMES_VARS['deposition_kgS/ha']    = ['wetso4']

    #: Dictionary mapping filenames to available variables in the respective files.
    FILES_CONTAIN = {}
    FILES_CONTAIN['monthly_so2.csv']        = ['concso2']
    FILES_CONTAIN['monthly_so4_aero.csv']   = ['concso4']
    FILES_CONTAIN['monthly_so4_precip.csv'] = ['wetso4', 'pr', 'concso4pr']

    #: Dictionary mapping variable name to hard coded filenames.
    VARS_TO_FILES = {}
    VARS_TO_FILES['concso2']   = ['monthly_so2.csv']
    VARS_TO_FILES['concso4']   = ['monthly_so4_aero.csv']
    VARS_TO_FILES['pr']        = ['monthly_so4_precip.csv']
    VARS_TO_FILES['wetso4']    = ['monthly_so4_precip.csv']
    VARS_TO_FILES['concso4pr'] = ['monthly_so4_precip.csv']

    # (from to unit)
    UNITCONVERSION = {'concso2':   ('ug S/m3', 'ug m-3'),
                      'concso4':   ('ug S/m3', 'ug m-3'),
                      'wetso4':    ('kg S/ha', 'kg m-2'),  #  s-1
                      'concso4pr': ('mg S/L',   'g m-3')
                      }

# =============================================================================
#     VAR_UNITS_READ={}
#     VAR_UNITS_READ['concso2'] = 'ug m-3'
#     VAR_UNITS_READ['concso4'] = 'ug m-3'
#     VAR_UNITS_READ['pr']       = 'mm'
#     VAR_UNITS_READ['wetso4']   = 'kg m-2 s-1'
#     VAR_UNITS_READ['concso4pr'] = 'g m-3' # Removed sulphur from VAR_UNITS_READ mgS/L
# =============================================================================
    #: :obj: `list` of :obj: `str`
    #: List containing all the variables available in this data set.
    PROVIDES_VARIABLES = list(VARS_TO_FILES.keys())

    #: int: Number of available variables in this data set.
    num_vars = len(PROVIDES_VARIABLES)

    @property
    def DEFAULT_VARS(self):
        """Default variables (wrapper for :attr:`PROVIDES_VARIABLES`)"""
        return self.PROVIDES_VARIABLES

    def _get_time_stamps(self, df):
        tconv = lambda yr, m : np.datetime64('{:04d}-{:02d}-{:02d}'.format(yr, m, 1), 's')
        dates_alt = [tconv(yr, m) for yr, m in
                     zip(df.year.values, df.month.values)]
        return np.asarray(dates_alt)

    def read_file(self, filename, vars_to_retrieve): #  -> List[StationData]:
        """ Read one GawTadSubsetAasEtAl file

        ToDo
        ----
        Could be more efficient...

        Note
        ----
        One file contains all stations for one variable. This is unctypical
        for ground based observations since the data files are typically by
        station.

        Parameters
        ----------
        filename : str
            absolute path to filename to read

        vars_to_retrieve : list or str, optional
            list containing variable IDs that are supposed to be read.

        Returns
        -------
        station_list : list
            list of `StationData` objects containing data
        """
        station_list = []
        df = pd.read_csv(filename, sep=",", low_memory=False)
        # Converting month and year.

        df['dtime'] = self._get_time_stamps(df)

        # array av numpy.datetime64
        df.rename(columns= {"Sampler":"instrument_name"}, inplace=True)
        grouped = df.groupby(by = "station_name")

        # Looping over every station:
        for name, station_group in grouped:
            station_group = station_group.drop_duplicates(subset='dtime',
                                                          keep='first')  # Drops duplacate rows

            tvals = station_group['dtime']

            stat = StationData()
            # Meta data
            stat['station_name'] = name

            stat["filename"] = filename
            stat["data_id"] = self.data_id
            stat["ts_type"] = self.TS_TYPE

            var_names = self.COLNAMES_VARS
            # Looping over all keys in the grouped data frame.
            for key in station_group:
                # Enters if the the key is a variable
                if key in var_names:
                    _var = np.intersect1d(var_names[key], vars_to_retrieve)
                    if len(_var) == 0:
                        continue
                    elif len(_var) > 1:
                        raise IOError('Found multiple matches...')
                    var = _var[0]
                    if var in self.UNITCONVERSION.keys():
                        # Convert units
                        from_unit, to_unit = self.UNITCONVERSION[var]
                        values = pd.to_numeric(station_group[key],
                                               errors='coerce').values
                        stat[var] = convert_unit(data=values,
                                                 from_unit=from_unit,
                                                 to_unit=to_unit,
                                                 var_name=var)

                        if var == 'wetso4':
                            # TODO: quite hard coded...
                            numsecs = get_tot_number_of_seconds(ts_type=self.TS_TYPE,
                                                                dtime=tvals)
                            stat[var] = stat[var] / numsecs

                            to_unit = 'kg m-2 s-1'
                    else:
                        to_unit = key.split('_')[-1]
                        vals = pd.to_numeric(station_group[key],
                                             errors='coerce').values
                        # This should only be true for
                        stat[var] = vals

                    stat['var_info'][var] = {'units' : to_unit}
                else:
                    if key == 'dtime':
                        stat[key] = station_group[key].values
                    else:
                        # Store the meta data.
                        stat[key] = station_group[key].values[0]

# =============================================================================
#             # Since no altitude is provided in the files. Geodesy is used to approximate
#             # altitude based on latitude and longitude.
#             try:
#                 s["altitude"] = geodesy.get_topo_altitude(lat=s.latitude,
#                                                           lon=s.longitude)
#             except ValueError as e:
#                 s['altitude'] = np.nan
#                 from pyaerocom import const
#                 const.logger.warning('Failed to access altitude for {}'
#                                      .format(name))
# =============================================================================
            # Added the created station to the station list.
            station_list.append(stat)
        return station_list

    def read(self, vars_to_retrieve=None):
        """ Method that reads list of files as instance of :class:`UngriddedData`

        Parameters
        ----------
        vars_to_retrieve : :obj:`list` or `str`:, optional,
            list containing variable IDs that are supposed to be read. If None,
            all variables in :attr:`PROVIDES_VARIABLES` are loaded

        files : :obj:`list`, optional
            list of files to be read. If None, then the file list is used that
            is returned on :func:`get_file_list`.

        Returns
        -------
        UngriddedData : :class:`UngriddedData`
            data object

        """

        files = self.get_file_list()
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
        data_obj = UngriddedData()

        meta_key = 0.0
        idx = 0
        varindex = -1

        #assign metadata object
        metadata = data_obj.metadata # OrderedDict
        meta_idx = data_obj.meta_idx # OrderedDict

        for file in files:
            filename = os.path.basename(file)
            if not filename in self.FILES_CONTAIN:
                raise IOError('Invalid file name {}, this should not happen.'
                              .format(filename))
            var_matches = [var for var in vars_to_retrieve if var in
                            self.FILES_CONTAIN[filename]]
            if len(var_matches) == 0:
                continue
            stat_list = self.read_file(file, vars_to_retrieve=var_matches)
            for stat in stat_list:
                #self.counter += 1
                metadata[meta_key] = OrderedDict()
                metadata[meta_key].update(stat.get_meta())
                metadata[meta_key].update(stat.get_station_coords())
                metadata[meta_key]['data_id'] = self.data_id
                metadata[meta_key]['ts_type'] = self.TS_TYPE
                #metadata[meta_key]['variables'] = stat["variables"]
                # Is instrumentname
                if 'instrument_name' in stat and stat['instrument_name'] is not None:
                    instr = stat['instrument_name']
                else:
                    instr = self.INSTRUMENT_NAME

                metadata[meta_key]['instrument_name'] = instr
                metadata[meta_key]['data_revision'] = self.data_revision

                # this is a list with indices of this station for each variable
                # not sure yet, if we really need that or if it speeds up things
                meta_idx[meta_key] = OrderedDict()

                num_times = len(stat['dtime'])
                num_vars = len(stat["var_info"])
                temp_vars = list(stat["var_info"].keys())
                tconv = stat['dtime'].astype('datetime64[s]')
                times = np.float64(tconv)
                totnum = num_times * num_vars

                if (idx + totnum) >= data_obj._ROWNO:
                    # This results in a error because it doesn't want to multiply empty with nan
                    data_obj.add_chunk(totnum)

                metadata[meta_key]['var_info'] = OrderedDict()
                for var_count, var in enumerate(temp_vars):

                    values = stat[var]
                    start = idx + var_count * num_times
                    stop = start + num_times

                    if not var in data_obj.var_idx:
                        varindex += 1
                        data_obj.var_idx[var] = varindex
                        var_idx = varindex
                    else:
                        var_idx = data_obj.var_idx[var]

                    metadata[meta_key]['var_info'][var] = stat['var_info'][var]

                    data_obj._data[start:stop, data_obj._LATINDEX] = stat['latitude']
                    data_obj._data[start:stop, data_obj._LONINDEX] = stat['longitude']
                    data_obj._data[start:stop, data_obj._ALTITUDEINDEX] = stat['altitude']
                    data_obj._data[start:stop, data_obj._METADATAKEYINDEX] = meta_key

                    # write data to data object
                    data_obj._data[start:stop, data_obj._TIMEINDEX] = times
                    data_obj._data[start:stop, data_obj._DATAINDEX] = values
                    data_obj._data[start:stop, data_obj._VARINDEX] = var_idx

                    meta_idx[meta_key][var] = np.arange(start, stop)

                meta_key += 1
                idx += totnum

        data_obj._data = data_obj._data[:idx]
        # sanity check
        data_obj._check_index()
        self.data = data_obj # initalizing a pointer to it selves
        return data_obj

def _check_line_endings(filename):
    """File that checks all line endings in an ascii file (not in use)"""
    ll = None
    wrong_endings = {}
    with open(filename, 'r') as f:
        prev = None
        for i, line in enumerate(f.readlines()):
            spl = line.split(',')
            if not spl[-1] == '\n':
                wrong_endings[i+1] = line
            if ll is None:
                ll = len(spl)
            elif not len(spl) == ll:
                print(i)
                print(prev)
                print(spl)
                raise Exception
            prev = spl
    return wrong_endings

class ReadSulphurAasEtAl(ReadAasEtal):
    """Old name of :class:`ReadAasEtal`."""

    def __init__(self, *args, **kwargs):
        super(ReadAasEtal, self).__init__(*args, **kwargs)
        msg = ('You are using an old name for class ReadAasEtal')
        const.print_log.warning(DeprecationWarning(msg))

if __name__ == "__main__":

     aa = ReadAasEtal('GAWTADsubsetAasEtAl')
     print(aa.data_id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May 27 12:27:47 2021

@author: jonasg
"""
from fnmatch import fnmatch
from pyaerocom._lowlevel_helpers import BrowseDict
from pyaerocom.exceptions import EntryNotAvailable, EvalEntryNameError
from pyaerocom.aeroval.obsentry import ObsEntry
from pyaerocom.aeroval.modelentry import ModelEntry

import abc


class BaseCollection(BrowseDict, abc.ABC):
    #: maximum length of entry names
    MAXLEN_KEYS = 25
    #: Invalid chars in entry names
    FORBIDDEN_CHARS_KEYS =['_']

    def _check_entry_name(self, key):
        if any([x in key for x in self.FORBIDDEN_CHARS_KEYS]):
            raise EvalEntryNameError(
                f'Invalid name: {key}. Must not contain any of the following '
                f'characters: {self.FORBIDDEN_CHARS_KEYS}')

    def __setitem__(self, key, value):
        self._check_entry_name(key)
        if 'web_interface_name' in value:
            self._check_entry_name(value['web_interface_name'])
        super().__setitem__(key, value)

    def keylist(self, name_or_pattern:str=None) -> list:
        """Find model names that match input search pattern(s)

        Parameters
        ----------
        name_or_pattern : str, optional
            Name or pattern specifying search string.

        Returns
        -------
        list
            list of keys in collection that match input requirements. If
            `name_or_pattern` is None, all keys will be returned.

        Raises
        ------
        KeyError
            if no matches can be found
        """
        if name_or_pattern is None:
            name_or_pattern = '*'

        matches = []
        for key in self.keys():
            if fnmatch(key, name_or_pattern) and not key in matches:
                matches.append(key)
        if len(matches) == 0:
            raise KeyError(
                f'No matches could be found that match input {name_or_pattern}')
        return matches

    @abc.abstractmethod
    def get_entry(self, key) -> object:
        """
        Getter for eval entries

        Raises
        ------
        KeyError
            if input name is not in this collection
        """
        pass

    @property
    @abc.abstractmethod
    def web_iface_names(self) -> list:
        """
        List of webinterface names for
        """
        pass

class ObsCollection(BaseCollection):
    """
    Dict-like object that represents a collection of obs entries

    Keys are obs names, values are instances of :class:`ObsEntry`. Values can
    also be assigned as dict and will automatically be converted into
    instances of :class:`ObsEntry`.


    Note
    ----
    Entries must not necessarily be only observations but may also be models.
    Entries provided in this collection refer to the y-axis in the AeroVal
    heatmap display and must fulfill the protocol defined by :class:`ObsEntry`.

    """
    SETTER_CONVERT = {dict : ObsEntry}

    def get_entry(self, key) -> object:
        """
        Getter for obs entries

        Raises
        ------
        KeyError
            if input name is not in this collection
        """
        try:
            entry = self[key]
            entry['obs_name'] = self.get_web_iface_name(key)
            return entry
        except (KeyError, AttributeError):
            raise EntryNotAvailable(f'no such entry {key}')

    def get_all_vars(self) -> list:
        """
        Get unique list of all obs variables from all entries

        Returns
        -------
        list
            list of variables specified in obs collection

        """
        vars = []
        for ocfg in self.values():
            vars.extend(ocfg.get_all_vars())
        return sorted(list(set(vars)))

    def get_web_iface_name(self, key):
        """
        Get webinterface name for entry

        Note
        ----
        Normally this is the key of the obsentry in :attr:`obs_config`,
        however, it might be specified explicitly via key `web_interface_name`
        in the corresponding value.

        Parameters
        ----------
        key : str
            key of entry.

        Returns
        -------
        str
            corresponding name

        """
        entry = self[key]
        if not 'web_interface_name' in entry:
            return key
        return entry['web_interface_name']

    @property
    def web_iface_names(self) -> list:
        """
        List of web interface names for each obs entry

        Returns
        -------
        list
        """
        return [self.get_web_iface_name(key) for key in self.keylist()]

    @property
    def all_vert_types(self):
        """List of unique vertical types specified in this collection"""
        return list(set([x['obs_vert_type'] for x in self.values()]))

class ModelCollection(BaseCollection):
    """
    Dict-like object that represents a collection of model entries

    Keys are model names, values are instances of :class:`ModelEntry`. Values
    can also be assigned as dict and will automatically be converted into
    instances of :class:`ModelEntry`.


    Note
    ----
    Entries must not necessarily be only models but may also be observations.
    Entries provided in this collection refer to the x-axis in the AeroVal
    heatmap display and must fulfill the protocol defined by
    :class:`ModelEntry`.

    """

    SETTER_CONVERT = {dict : ModelEntry}
    def get_entry(self, key) -> object:
        """Get model entry configuration

        Since the configuration files for experiments are in json format, they
        do not allow the storage of executable custom methods for model data
        reading. Instead, these can be specified in a python module that may
        be specified via :attr:`add_methods_file` and that contains a
        dictionary `FUNS` that maps the method names with the callable methods.

        As a result, this means that, by default, custom read methods for
        individual models in :attr:`model_config` do not contain the
        callable methods but only the names. This method will take care of
        handling this and will return a dictionary where potential custom
        method strings have been converted to the corresponding callable
        methods.

        Parameters
        ----------
        model_name : str
            name of model

        Returns
        -------
        dict
            Dictionary that specifies the model setup ready for the analysis
        """
        try:
            entry = self[key]
            entry['model_name'] = key
            return entry
        except (KeyError,AttributeError):
            raise EntryNotAvailable(f'no such entry {key}')

    @property
    def web_iface_names(self) -> list:
        """
        List of web interface names for each obs entry

        Returns
        -------
        list
        """
        return self.keylist()


import os
from configparser import ConfigParser

from pyaerocom import const
from pyaerocom._lowlevel_helpers import BrowseDict
from pyaerocom.io import EbasSQLRequest
from pyaerocom.exceptions import VarNotAvailableError

class EbasVarInfo(BrowseDict):
    """Interface for mapping between EBAS variable information and AeroCom

    For more information about EBAS variable and data information see
    `EBAS website <http://ebas.nilu.no/>`__.

    Attributes
    ----------
    var_name : str
        AeroCom variable name
    component : list
        list of EBAS variable / component names that are mapped to
        :attr:`var_name`
    matrix : list, optional
        list of EBAS matrix values that are accepted, default is None, i.e.
        all available matrices are used
    instrument : list, optional
        list of all instruments that are accepted for this variable
    requires : list, optional
        for variables that are computed and not directly available in EBAS.
        Provided as list of (AeroCom) variables that are required to
        compute :attr:`var_name` (e.g. for `sc550dryaer` this would be
        `[sc550aer,scrh]`).
    scale_factor : float, optional
        multiplicative scale factor that is applied in order to convert
        EBAS variable into AeroCom variable (e.g. 1.4 for conversion of
        EBAS OC measurement to AeroCom concoa variable)

    Parameters
    ----------
    var_name : str
        AeroCom variable name
    init : bool
        if True, EBAS configuration for input variable is retrieved from
        data file ebas_config.ini (if possible)
    **kwargs
        additional keyword arguments (currently not used)

    """

    def __init__(self, var_name, init=True, **kwargs):
        self.var_name = var_name

        self.component = None

        #: list of matrix names (EBAS side, optional)
        self.matrix = None

        #: list of instrument names (EBAS side, optional)
        self.instrument = None

        #: list containing variable statistics info (EBAS side, optional)
        self.statistics = None

        #: list of additional variable required for retrieval of this variable
        self.requires = None

        #: scale factor for conversion to Aerocom units
        self.scale_factor = 1

        #imports default information and, on top, variable information (if
        # applicable)
        if init:
            self.parse_from_ini(var_name)

    @staticmethod
    def PROVIDES_VARIABLES():
        """List specifying provided variables"""
        data = EbasVarInfo.open_config()
        return [k for k in data.keys()]

    @staticmethod
    def open_config():
        """Open ebas_config.ini file with `ConfigParser`

        Returns
        -------
        ConfigParser
        """
        from pyaerocom import __dir__
        fpath = os.path.join(__dir__, "data", "ebas_config.ini")
        conf_reader = ConfigParser()
        conf_reader.read(fpath)
        return conf_reader

    @property
    def var_name_aerocom(self):
        """Variable name in AeroCom convention"""
        return const.VARS[self.var_name].var_name_aerocom

    def parse_from_ini(self, var_name=None, conf_reader=None):
        """
        Parse EBAS info for input AeroCom variable (works also for aliases)

        Parameters
        ----------
        var_name : str
            AeroCom variable name
        conf_reader : ConfigParser
            open config parser object

        Raises
        ------
        VarNotAvailableError
            if variable is not supported

        Returns
        -------
        bool
            True, if default could be loaded, False if not
        """
        if conf_reader is None:
            conf_reader = self.open_config()

        if not var_name in conf_reader:
            # this will raise Variable
            var_name = const.VARS[var_name].var_name_aerocom
            if not var_name in conf_reader:
                raise VarNotAvailableError('Variable {} is not available in '
                                           'EBAS interface'.format(var_name))

        var_info = conf_reader[var_name]
        for key in self.keys():
            if key in var_info:
                val = var_info[key]
                if key == 'scale_factor':
                    self[key] = float(val.split('#')[0].strip())
                else:
                    self[key] = list(dict.fromkeys([x for x in val.split(',')]))
        self.var_name=var_name

    def to_dict(self):
        """Convert into dictionary"""
        d = {}
        for k, v in self.items():
            if k == 'unit':
                k = 'units'
            if v is not None:
                d[k] = v
        return d

    def make_sql_request(self, **constraints):
        """Create an SQL request for the specifications in this object

        Parameters
        ----------
        constraints
            request constraints deviating from default. For details on
            parameters see :class:`EbasSQLRequest`

        Returns
        -------
        EbasSQLRequest
            the SQL request object that can be used to retrieve corresponding
            file names using instance of :func:`EbasFileIndex.get_file_names`.
        """
        if self.requires is not None:
            raise ValueError(
                f'This variable {self.var_name} requires other variables '
                f'for reading, thus more than one SQL request is needed. '
                f'Please use :func:`make_sql_requests` instead')

        variables = self.component

        if variables is None:
            raise AttributeError(
                f'At least one component (Ebas variable name) '
                f'must be specified for retrieval of variable {self.var_name}'
                )

        # default request
        req = EbasSQLRequest(variables=variables, matrices=self.matrix,
                              instrument_types=self.instrument,
                              statistics=self.statistics)

        req.update(**constraints)
        return req

    def make_sql_requests(self, **constraints):
        """Create a list of SQL requests for the specifications in this object

        Parameters
        ----------
        requests : dict, optional
            other SQL requests linked to this one (e.g. if this variable
            requires)
        constraints
            request constraints deviating from default. For details on
            parameters see :class:`EbasSQLRequest`

        Returns
        -------
        list
            list of :class:`EbasSQLRequest` instances for this component and
            potential required components.
        """
        requests = {}
        if self.component is not None:
            req = EbasSQLRequest(variables=self.component,
                                 matrices=self.matrix,
                                 instrument_types=self.instrument,
                                 statistics=self.statistics)
            req.update(**constraints)
            requests[self.var_name] = req

        if self.requires is not None:
            for var in self.requires:
                if var in requests:
                    # ToDo: check if this can be generalised better
                    raise ValueError(
                        f'Variable conflict in EBAS SQL request: '
                        f'{var} cannot depent on itself...')
                info = EbasVarInfo(var)
                _reqs = info.make_sql_requests(**constraints)
                for _var, _req in _reqs.items():
                    if _var in requests:
                        # ToDo: check if this can be generalised better
                        raise ValueError(
                            f'Variable conflict in EBAS SQL request: '
                            f'{_var} cannot depent on itself...')
                    requests[_var] = _req

        return requests

    def __str__(self):
        head = "Pyaerocom {}".format(type(self).__name__)
        s = "\n{}\n{}".format(head, len(head)*"-")
        for k, v in self.items():
                s += "\n%s: %s" %(k,v)
        return s



import fnmatch
import os
from configparser import ConfigParser

from cf_units import Unit

from pyaerocom import logger
from pyaerocom.variable import Variable
from pyaerocom.exceptions import VariableDefinitionError
from pyaerocom.variable_helpers import parse_variables_ini, parse_aliases_ini


class VarCollection(object):
    """Variable access class based on variables.ini file"""

    def __init__(self, var_ini):
        self._all_vars = None
        self._var_ini = None

        self.var_ini = var_ini

        self._vars_added = {}

        self._cfg_parser = parse_variables_ini(var_ini)
        self._alias_parser = parse_aliases_ini()
        self._idx = -1

    @property
    def all_vars(self):
        """List of all variables

        Note
        ----
        Does not include variable names that may be inferred via
        alias families as defined in section [alias_families] in
        aliases.ini.
        """
        if self._all_vars is None:
            all_vars = [k for k in self._cfg_parser.keys()]
            all_vars.extend(self._vars_added.keys())
            self._all_vars = all_vars
        return self._all_vars


    @property
    def var_ini(self):
        """Config file specifying variable information"""
        return self._var_ini


    @var_ini.setter
    def var_ini(self, var_ini):
        if not isinstance(var_ini, str):
            raise ValueError(f'Invalid input for var_ini, need str')
        elif not os.path.exists(var_ini):
            raise FileNotFoundError(f'File {var_ini} does not exist')
        self._var_ini = var_ini


    def add_var(self, var):
        """Add a new variable to this collection

        Minimum requirement for new variables are attributes var_name and
        units.

        Parameters
        ----------
        var : Variable
            new variable definition

        Raises
        ------
        VariableDefinitionError
            if a variable is already defined under that name

        Returns
        -------
        None
        """
        if not isinstance(var.var_name, str):
            raise ValueError('Attr. var_name needs to be assigned to input '
                             'variable')
        if var.var_name in self.all_vars:
            raise VariableDefinitionError(f'variable with name {var.var_name} '
                                          f'is already defined')
        if not isinstance(var, Variable):
            raise ValueError('Can only add instances of Variable class...')
        if not isinstance(var.units, str):
            if not isinstance(var.units, Unit):
                raise ValueError('Please assign a unit to the new input '
                                 'variable')
            var.units = str(var.units)
        self._all_vars.append(var.var_name)
        self._vars_added[var.var_name] = var


    def delete_variable(self, var_name:str) -> None:
        """
        Remove input variable from this collection

        Parameters
        ----------
        var_name : str
            name of variable

        Raises
        ------
        VariableDefinitionError
            if variable does not exist or if it exists more than once.

        Returns
        -------
        None

        """
        all_vars = self.all_vars
        matches = [i for i, x in enumerate(all_vars) if x==var_name]
        if len(matches) == 0:
            raise VariableDefinitionError(
                f'No such variable {var_name} in VarCollection')
        elif len(matches) >1:
            raise VariableDefinitionError(
                f'FATAL: found multiple matches for variable {var_name} in '
                f'VarCollection. Please check variables.ini')
        all_vars.pop(matches[0])
        self._all_vars == all_vars
        if var_name in self._vars_added:
            del self._vars_added[var_name]

    def get_var(self, var_name):
        """
        Get variable based on variable name

        Parameters
        ----------
        var_name : str
            name of variable

        Raises
        ------
        VariableDefinitionError
            if no variable under input var_name is registered.

        Returns
        -------
        Variable
            Variable instance

        """
        if var_name in self._vars_added:
            return self._vars_added[var_name]
        var = Variable(var_name, cfg=self._cfg_parser)
        if not var.var_name_aerocom in self:
            raise VariableDefinitionError(
                'Error (VarCollection): input variable '
                '{} is not supported'.format(var_name))
        return var

    def find(self, search_pattern):
        """Find all variables that match input search pattern

        Note
        ----
        Searches for matches in variable names (:attr:`Variable.var_name`) and
        standard name (:attr:`Variable.standard_name`).

        Parameters
        ----------
        search_pattern : str
            variable search pattern

        Returns
        -------
        list
            AeroCom variable names that match the search pattern
        """
        matches = []
        for var in self:
            if fnmatch.fnmatch(var.var_name, search_pattern):
                matches.append(var.var_name)
            elif (isinstance(var.standard_name, str) and
                  fnmatch.fnmatch(var.standard_name, search_pattern)):
                matches.append(var.var_name)
        return matches

    def __dir__(self):
        """Activates auto tab-completion for all variables"""
        return self.all_vars

    def __iter__(self):
        return self

    def __next__(self):
        self._idx += 1
        if self._idx == len(self.all_vars):
            self._idx = -1
            raise StopIteration
        var_name = self.all_vars[self._idx]
        return self[var_name]

    def __contains__(self, var_name):
        if var_name in self.all_vars:
            return True
        return False

    def __len__(self):
        return len(self.all_vars)

    def __getitem__(self, var_name):
        return self.get_var(var_name)

    def __repr__(self):
        return f'VarCollection ({len(self)} entries)'

    def __str__(self):
        return repr(self)


def browse_database(*args, **kwargs):
    from pyaerocom.io.utils import browse_database
    return browse_database(*args, **kwargs)

def clear_cache():
    """
    Delete all *.pkl files in cache directory
    """
    from pyaerocom.io.cachehandler_ungridded import CacheHandlerUngridded
    ch = CacheHandlerUngridded()
    ch.delete_all_cache_files()


from geonum.atmosphere import p0, T0_STD
from pyaerocom.aux_var_helpers import concx_to_vmrx
from pyaerocom.molmasses import get_molmass
from pyaerocom.units_helpers import get_unit_conversion_fac


def add_dataarrays(*arrs):
    """
    Add a bunch of :class:`xarray.DataArray` instances

    Parameters
    ----------
    *arrs
        input arrays (instances of :class:`xarray.DataArray` with same shape)

    Returns
    -------
    xarray.DataArray
        Added array

    """
    if not len(arrs) > 1:
        raise ValueError('Need at least 2 input arrays to add')
    result = arrs[0].copy(deep=True)
    for arr in arrs[1:]:
        result += arr
    return result


def subtract_dataarrays(*arrs):
    """
    Subtract a bunch of :class:`xarray.DataArray` instances from an array

    Parameters
    ----------
    *arrs
        input arrays (instances of :class:`xarray.DataArray` with same shape).
        Subtraction is performed with respect to the first input array.


    Returns
    -------
    xarray.DataArray
        Diff array (all additional ones are subtracted from first array)

    """
    if not len(arrs) > 1:
        raise ValueError('Need at least 2 input arrays to add')
    result = arrs[0].copy(deep=True)
    for arr in arrs[1:]:
        result -= arr
    return result


def calc_concNhno3(*arrs):
    if len(arrs)>1:
        raise ValueError('Shoul only be given 1 array')

    M_N = 14.006
    M_O = 15.999
    M_H = 1.007

    conchno3 = arrs[0].copy(deep=True)
    fac = (M_N / (M_H + M_N + M_O * 3))
    concNhno3 = conchno3*fac
    concNhno3.attrs['units'] = 'ug N m-3'
    return concNhno3

# ToDo: add docstring
def calc_concNno3pm10(concno3f,concno3c):
    M_N = 14.006
    M_O = 15.999

    fac = M_N / (M_N + 3*M_O)
    concno3pm10 = concno3f + concno3c
    concNno3pm10 = concno3pm10*fac
    concNno3pm10.attrs['var_name'] = 'concNno3pm10'
    concNno3pm10.attrs['units'] = 'ug N m-3'
    return concNno3pm10

# ToDo: add docstring
def calc_concNno3pm25(concno3f,concno3c,fine_from_coarse_fraction=0.134):
    M_N = 14.006
    M_O = 15.999

    fac = M_N / (M_N + 3*M_O)
    concno3pm25 = concno3f + fine_from_coarse_fraction*concno3c
    concNno3pm25 = concno3pm25*fac
    concNno3pm25.attrs['var_name'] = 'concNno3pm25'
    concNno3pm25.attrs['units'] = 'ug N m-3'
    return concNno3pm25

# ToDo: add docstring
def calc_conNtno3(conchno3,concno3f,concno3c):
    concNhno3 = calc_concNhno3(conchno3)
    concNno3pm10 = calc_concNno3pm10(concno3f,concno3c)

    concNtno3 = concNhno3 + concNno3pm10
    concNtno3.attrs['units'] = 'ug N m-3'
    return concNtno3

# ToDo: add docstring
def calc_concNnh3(*arrs):
    if len(arrs)>1:
        raise ValueError('Shoul only be given 1 array')

    M_N = 14.006
    M_H = 1.007

    concnh3 = arrs[0].copy(deep=True)
    concNnh3 = concnh3*(M_N / (M_H * 3 + M_N))
    concNnh3.attrs['units'] = 'ug N m-3'
    return concNnh3

# ToDo: add docstring
def calc_concNnh4(*arrs):
    if len(arrs)>1:
        raise ValueError('Shoul only be given 1 array')

    M_N = 14.006
    M_H = 1.007

    concnh4 = arrs[0].copy(deep=True)
    concNnh4 = concnh4*(M_N / (M_H * 4 + M_N))
    concNnh4.attrs['units'] = 'ug N m-3'
    return concNnh4

# ToDo: add docstring
def calc_concNtnh(concnh3,concnh4):
    concNnh3 = calc_concNnh3(concnh3)
    concNnh4 = calc_concNnh4(concnh4)

    concNtnh = concNnh3 + concNnh4
    concNtnh.attrs['units'] = 'ug N m-3'
    return concNtnh

# ToDo: add docstring
def update_EC_units(concecpm25):
    concCecpm25 = concecpm25
    concCecpm25.attrs['units'] = 'ug C m-3'

    return concCecpm25


def calc_concsspm25(concssf, concssc, coarse_fraction=0.13):
    """
    Calculate PM2.5 seasalt

    Parameters
    ----------
    concssf : xr.DataArray
        EMEP output fine seasalt
    concssc : xr.DataArray
        EMEP output coarse seasalt
    coarse_fraction : float
        fraction of coarse supposed to be added to fine output, defaults to
        0.13.

    Returns
    -------
    xr.DataArray
        PM2.5 seasalt including input coarse fraction

    """
    concsspm25 = concssf + coarse_fraction*concssc

    concsspm25.attrs['units'] = 'ug m-3'
    return concsspm25


def calc_vmrox(concno2, vmro3):
    """
    Calculate OX VMR from NO2 concentration and O3 VMR

    Converts NO2 conc to NO2 VMR assuming US standard atmosphere and adds
    that with VMR O3.

    Parameters
    ----------
    concno2 : xr.DataArray
        mass concentration of NO2
    vmro3 : xr.DataArray
        volume mixing ratio of O3

    Returns
    -------
    xr.DataArray
        volume mixing ratio of OX (O3 + NO2) in units of nmole mole-1

    """
    o3mulfac = get_unit_conversion_fac(1, vmro3.attrs['units'], 'nmole mole-1')
    if o3mulfac != 1:
        vmro3 = vmro3.copy(deep=True)
        vmro3 *= o3mulfac
        vmro3.attrs['units'] = 'nmole mole-1'

    vmrno2 = concx_to_vmrx(
        data=concno2,
        p_pascal=p0, # 1013 hPa (US standard atm)
        T_kelvin=T0_STD, #15 deg celcius (US standard atm)
        conc_unit=str(concno2.attrs['units']),
        mmol_var=get_molmass('no2'), # g/mol NO2
        to_unit="nmol mol-1"
    )

    vmrox = vmrno2 + vmro3
    vmrox.attrs["units"] = "nmol mol-1"
    return vmrox


"""
Variable categorisation groups

These are needed in some cases to infer, e.g. units associated with variable
names. Used in :class:`pyaerocom.variable.Variable` to identify certain groups.

Note
-----
The below definitions are far from complete
"""

#: start string of emission variables
emi_startswith = 'emi'
#: start string of wet deposition variables
wetdep_startswith = 'wet'
#: start string of dry deposition variables
drydep_startswith = 'dry'
#: start string of total deposition variables
totdep_startswith = 'dep'

#: additional emission rate variables (that do not start with emi*)
emi_add_vars = []

#: additional deposition rate variables (that do not start with wet* or dry*)
dep_add_vars = []


import abc
import glob, os
import logging
import numpy as np
from fnmatch import fnmatch

from pyaerocom import const
from pyaerocom._lowlevel_helpers import list_to_shortstr
from pyaerocom.io.helpers import get_obsnetwork_dir
from pyaerocom import LOGLEVELS
from pyaerocom.helpers import varlist_aerocom
from pyaerocom.exceptions import DataSourceError
# TODO: Proposal: include attribute ts_type that is by default undefined but
# may be set to either of the defined
class ReadUngriddedBase(abc.ABC):
    """TEMPLATE: Abstract base class template for reading of ungridded data

    .. note::

        The two dictionaries ``AUX_REQUIRES`` and ``AUX_FUNS`` can be filled
        with variables that are not contained in the original data files but
        are computed during the reading. The former specifies what additional
        variables are required to perform the computation and the latter
        specifies functions used to perform the computations of the auxiliary
        variables.
        See, for instance, the class :class:`ReadAeronetSunV3`, which includes
        the computation of the AOD at 550nm and the Angstrom coefficient
        (in 440-870 nm range) from AODs measured at other wavelengths.
    """
    #: version of this base class. Please update if you apply changes to this
    #: code. This version is required for caching and needs to be considered
    #: in the definition of __version__ in all derived classes, so that
    #: caching can be done reliably
    __baseversion__ = '0.09'

    #: dictionary containing information about additionally required variables
    #: for each auxiliary variable (i.e. each variable that is not provided
    #: by the original data but computed on import)
    AUX_REQUIRES = {}

    #: Functions that are used to compute additional variables (i.e. one
    #: for each variable defined in AUX_REQUIRES)
    AUX_FUNS = {}

    IGNORE_META_KEYS = []

    _FILEMASK = '*.*'

    def __str__(self):
        return ("Dataset name: {}\n"
                "Data directory: {}\n"
                "Supported variables: {}\n"
                "Last revision: {}"
                .format(self.data_id, self.data_dir,
                        self.PROVIDES_VARIABLES, self.data_revision))
    def __repr__(self):
        return str(type(self).__name__)

    @abc.abstractproperty
    def TS_TYPE(self):
        """Temporal resolution of dataset

        This should be defined in the header of an implementation class if
        it can be globally defined for the corresponding obs-network or in
        other cases it should be initated as string ``undefined`` and then,
        if applicable, updated in the reading routine of a file.

        The TS_TYPE information should ultimately be written into the meta-data
        of objects returned by the implementation of :func:`read_file` (e.g.
        instance of :class:`StationData` or a normal dictionary) and the method
        :func:`read` (which should ALWAYS return an instance of the
        :class:`UngriddedData` class).

        Note
        ----
        - Please use ``"undefined"`` if the derived class is not sampled on \
            a regular basis.
        - If applicable please use Aerocom ts_type (i.e. hourly, 3hourly, \
                                                    daily, monthly, yearly)
        - Note also, that the ts_type in a derived class may or may not be \
            defined in a general case. For instance, in the EBAS database the \
            resolution code can be found in the file header and may thus be \
            intiated as ``"undefined"`` in the initiation of the reading class \
            and then updated when the class is being read
        - For derived implementation classes that support reading of multiple \
            network versions, you may also assign
        """
        pass

    @abc.abstractproperty
    def _FILEMASK(self):
        """Mask for identifying datafiles (e.g. '*.txt')

        Note
        ----
        May be implemented as global constant in header
        """
        pass

    @abc.abstractproperty
    def __version__(self):
        """Version of reading class

        Keeps track of changes in derived reading class (e.g. to assess whether
        potential cache-files are outdated).

        Note
        ----
        May be implemented as global constant in header
        """
        pass

    @abc.abstractproperty
    def DATA_ID(self):
        """Name of dataset (OBS_ID)

        Note
        ----

        - May be implemented as global constant in header of derieved class
        - May be multiple that can be specified on init (see example below)

        """
        pass

    @abc.abstractproperty
    def SUPPORTED_DATASETS(self):
        """List of all datasets supported by this interface

        Note
        ----

        - best practice to specify in header of class definition
        - needless to mention that :attr:`DATA_ID` needs to be in this list
        """
        pass

    @abc.abstractproperty
    def PROVIDES_VARIABLES(self):
        """List of variables that are provided by this dataset

        Note
        ----
        May be implemented as global constant in header
        """
        pass

    @abc.abstractproperty
    def DEFAULT_VARS(self):
        """List containing default variables to read"""
        pass

    @abc.abstractmethod
    def read_file(self, filename, vars_to_retrieve=None):
        """Read single file

        Parameters
        ----------
        filename : str
            string specifying filename
        vars_to_retrieve : :obj:`list` or similar, optional,
            list containing variable IDs that are supposed to be read. If None,
            all variables in :attr:`PROVIDES_VARIABLES` are loaded

        Returns
        -------
        :obj:`dict` or :obj:`StationData`, or other...
            imported data in a suitable format that can be handled by
            :func:`read` which is supposed to append the loaded results from
            this method (which reads one datafile) to an instance of
            :class:`UngriddedData` for all files.
        """
        pass

    @abc.abstractmethod
    def read(self, vars_to_retrieve=None, files=[], first_file=None,
             last_file=None):
        """Method that reads list of files as instance of :class:`UngriddedData`

        Parameters
        ----------
        vars_to_retrieve : :obj:`list` or similar, optional,
            list containing variable IDs that are supposed to be read. If None,
            all variables in :attr:`PROVIDES_VARIABLES` are loaded
        files : :obj:`list`, optional
            list of files to be read. If None, then the file list is used that
            is returned on :func:`get_file_list`.
        first_file : :obj:`int`, optional
            index of first file in file list to read. If None, the very first
            file in the list is used
        last_file : :obj:`int`, optional
            index of last file in list to read. If None, the very last file
            in the list is used

        Returns
        -------
        UngriddedData
            instance of ungridded data object containing data from all files.
        """
        pass

    ### Concrete implementations of methods that are the same for all (or most)
    # of the derived reading classes
    def __init__(self, data_id=None, data_dir=None):
        self.data = None #object that holds the loaded data
        self._data_id = None
        self.files = []
        # list that will be updated in read method to store all files that
        # could not be read. It is the responsibility of developers of derived
        # classes to include a try / except block in method read, where the
        # method read_file is called, and in case of an Exception, append the
        # corresponding file path to this list.
        self.read_failed = []

        self._data_dir = data_dir

        #: Class own instance of logger class
        self.logger = logging.getLogger(__name__)
        self._add_aux_variables()

        if data_id is not None:
            if not data_id in self.SUPPORTED_DATASETS:
                raise AttributeError("Dataset {} not supported by this "
                                     "interface".format(data_id))
            self._data_id = data_id

    @property
    def data_id(self):
        """ID of dataset"""
        return self.DATA_ID if self._data_id is None else self._data_id

    @property
    def DATASET_PATH(self):
        """Wrapper for :attr:`data_dir`.
        """
        const.print_log.warning(DeprecationWarning(
            'WARNING: Attr. DATASET_PATH is deprecated in ungridded readers '
            'as of pyaerocom v0.11.0. Please use data_dir instead.'

            ))
        return self.data_dir

    @property
    def data_dir(self):
        """
        str: Location of the dataset

        Note
        ----
        This can be set explicitly when instantiating the class (e.g. if data
        is available on local machine). If unspecified, the data location is
        attempted to be inferred via :func:`get_obsnetwork_dir`

        Raises
        ------
        FileNotFoundError
            if data directory does not exist or cannot be retrieved
            automatically
        """
        if self._data_dir is None:
            self._data_dir = get_obsnetwork_dir(self.data_id)
        if not os.path.exists(self._data_dir):
            raise (f'{self._data_dir} does not exist.')
        return self._data_dir

    @property
    def REVISION_FILE(self):
        """Name of revision file located in data directory"""
        return const.REVISION_FILE

    @property
    def AUX_VARS(self):
        """List of auxiliary variables (keys of attr. :attr:`AUX_REQUIRES`)

        Auxiliary variables are those that are not included in original files
        but are computed from other variables during import
        """
        return list(self.AUX_REQUIRES.keys())



    @property
    def data_revision(self):
        """Revision string from file Revision.txt in the main data directory
        """
        if '_data_revision' in self.__dict__:
            return self.__dict__['_data_revision']
        rev = 'n/d'
        try:
            revision_file = os.path.join(self.data_dir, self.REVISION_FILE)
            if os.path.isfile(revision_file):
                with open(revision_file, 'rt') as in_file:
                    rev = in_file.readline().strip()
        except Exception:
            pass
        self._data_revision = rev
        return rev

    @property
    def verbosity_level(self):
        """Current level of verbosity of logger"""
        return self.logger.level

    @verbosity_level.setter
    def verbosity_level(self, val):
        if isinstance(val, str):
            if not val in LOGLEVELS:
                raise ValueError("Invalid input for loglevel")
            val = LOGLEVELS[val]
        self.logger.setLevel(val)

    def _add_aux_variables(self):
        """Helper that makes sure all auxiliary variables can be computed"""
        for var in self.AUX_REQUIRES.keys():
            if not var in self.AUX_FUNS:
                raise AttributeError("Fatal: no computation method defined for "
                                     "auxiliary variable {}. Please specify "
                                     "method in class header dictionary "
                                     "AUX_FUNS".format(var))
            if not var in self.PROVIDES_VARIABLES:
                self.PROVIDES_VARIABLES.append(var)

    def _add_additional_vars(self, vars_to_retrieve):
        """Add required additional variables for computation to input list

        Helper method that is called in :func:`check_vars_to_retrieve`
        in order to find all variables that are required for a specified
        retrieval. This is relevant for additionally computed variables
        (attribute ``AUX_VARS``) that are not available in the original data
        files, but are computed from available parameters.

        Parameters
        ----------
        vars_to_retrieve : list
            list of variables supported by this interface (i.e. must be
            contained in ``PROVIDES_VARIABLES``)

        Returns
        -------
        tuple
            2-element tuple, containing

            - bool : boolean, specifying whether variables list of required \
            variables needs to be extended or the order was changed
            - list : additionally required variables
        """
        changed = False
        added_vars = []

        for var in vars_to_retrieve:
            if var in self.AUX_VARS:
                vars_req = self.AUX_REQUIRES[var]
                for var_req in vars_req:
                    if var_req in vars_to_retrieve:
                        idx_var = vars_to_retrieve.index(var)
                        idx_var_req = vars_to_retrieve.index(var_req)
                        if idx_var < idx_var_req: #wrong order for computation
                            vars_to_retrieve[idx_var] = var_req
                            vars_to_retrieve[idx_var_req] = var
                            # break and return that it was changed (i.e repeat
                            # calling this method until nothing is changed or
                            # added)
                            return (True, added_vars + vars_to_retrieve)
                    else:
                        added_vars.append(var_req)
                        changed = True
        # it is important to insert the additionally required variables in
        # the beginning, as these need to be computed first later on
        # Example: if vars_to_retrieve=['od550aer'] then this loop will
        # find out that this requires 'ang4487aer' to be computed as
        # well. So at the end of this function, ang4487aer needs to be
        # before od550aer in the list vars_to_compute, since the method
        # "compute_additional_vars" loops over that list in the specified
        # order
        vars_to_retrieve = added_vars + vars_to_retrieve
        return (changed, vars_to_retrieve)

    def var_supported(self, var_name):
        """
        Check if input variable is supported

        Parameters
        ----------
        var_name : str
            AeroCom variable name or alias

        Raises
        ------
        VariableDefinitionError
            if input variable is not supported by pyaerocom

        Returns
        -------
        bool
            True, if variable is supported by this interface, else False

        """
        if (var_name in self.PROVIDES_VARIABLES or
            const.VARS[var_name].var_name_aerocom in self.PROVIDES_VARIABLES):
            return True
        return False

    def check_vars_to_retrieve(self, vars_to_retrieve):
        """Separate variables that are in file from those that are computed

        Some of the provided variables by this interface are not included in
        the data files but are computed within this class during data import
        (e.g. od550aer, ang4487aer).

        The latter may require additional parameters to be retrieved from the
        file, which is specified in the class header (cf. attribute
        ``AUX_REQUIRES``).

        This function checks the input list that specifies all required
        variables and separates them into two lists, one that includes all
        variables that can be read from the files and a second list that
        specifies all variables that are computed in this class.

        Parameters
        ----------
        vars_to_retrieve : list
            all parameter names that are supposed to be loaded

        Returns
        -------
        tuple
            2-element tuple, containing

            - list: list containing all variables to be read
            - list: list containing all variables to be computed
        """
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
        # first, check if input variables are alias names, and replace
        vars_to_retrieve = varlist_aerocom(vars_to_retrieve)

        repeat = True
        while repeat:
            repeat, vars_to_retrieve = self._add_additional_vars(vars_to_retrieve)

        # unique list containing all variables that are supposed to be read,
        # either because they are required to be retrieved, or because they
        # are supposed to be read because they are required to compute one
        # of the output variables
        vars_to_retrieve = list(dict.fromkeys(vars_to_retrieve))

        # in the following, vars_to_retrieve is separated into two arrays, one
        # containing all variables that can be read from the files, and the
        # second containing all variables that are computed
        vars_to_read = []
        vars_to_compute = []

        for var in vars_to_retrieve:
            if not var in self.PROVIDES_VARIABLES:
                raise ValueError("Invalid variable {}".format(var))
            elif var in self.AUX_REQUIRES:
                vars_to_compute.append(var)
            else:
                vars_to_read.append(var)
        return (vars_to_read, vars_to_compute)

    def compute_additional_vars(self, data, vars_to_compute):
        """Compute all additional variables

        The computations for each additional parameter are done using the
        specified methods in ``AUX_FUNS``.

        Parameters
        ----------
        data : dict-like
            data object containing data vectors for variables that are required
            for computation (cf. input param ``vars_to_compute``)
        vars_to_compute : list
            list of variable names that are supposed to be computed.
            Variables that are required for the computation of the variables
            need to be specified in :attr:`AUX_VARS` and need to be
            available as data vectors in the provided data dictionary (key is
            the corresponding variable name of the required variable).

        Returns
        -------
        dict
            updated data object now containing also computed variables
        """
        if not 'var_info' in data:
            data['var_info'] = {}
        for var in vars_to_compute:
            required = self.AUX_REQUIRES[var]
            missing = []
            for req in required:
                if not req in data:
                    missing.append(req)

            if len(missing) == 0:
                data[var] = self.AUX_FUNS[var](data)
                try:
                    data['var_info'][var]['computed']=True
                except KeyError:
                    data['var_info'][var] = {'computed' : True}

        return data

    def remove_outliers(self, data, vars_to_retrieve, **valid_rng_vars):
        """Remove outliers from data

        Parameters
        ----------
        data : dict-like
            data object containing data vectors for variables that are required
            for computation (cf. input param ``vars_to_compute``)
        vars_to_retrieve : list
            list of variable names for which outliers will be removed from
            data
        **valid_rng_vars
            additional keyword args specifying variable name and corresponding
            min / max interval (list or tuple) that specifies valid range
            for the variable. For each variable that is not explicitely defined
            here, the default minimum / maximum value is used (accessed via
            ``pyaerocom.const.VARS[var_name]``)
        """
        for var in vars_to_retrieve:
            if var in data:
                if var in valid_rng_vars:
                    rng = valid_rng_vars[var]
                    low, high =  rng[0], rng[1]
                else:
                    var_info = const.VARS[var]
                    low, high = var_info['minimum'], var_info['maximum']
                vals = data[var]
                mask = np.logical_or(vals < low, vals > high)
                vals[mask] = np.nan
                data[var] = vals
        return data

    def find_in_file_list(self, pattern=None):
        """Find all files that match a certain wildcard pattern

        Parameters
        ----------
        pattern : :obj:`str`, optional
            wildcard pattern that may be used to narrow down the search (e.g.
            use ``pattern=*Berlin*`` to find only files that contain Berlin
            in their filename)

        Returns
        -------
        list
            list containing all files in :attr:`files` that match pattern

        Raises
        ------
        IOError
            if no matches can be found
        """
        if len(self.files) == 0:
            self.get_file_list()
        files = [f for f in self.files if fnmatch(f, pattern)]
        if not len(files) > 0:
            raise IOError(f'No files could be detected that match the '
                          f'pattern {pattern}')
        return files

    def get_file_list(self, pattern=None):
        """Search all files to be read

        Uses :attr:`_FILEMASK` (+ optional input search pattern, e.g.
        station_name) to find valid files for query.

        Parameters
        ----------
        pattern : str, optional
            file name pattern applied to search

        Returns
        -------
        list
            list containing retrieved file locations

        Raises
        ------
        IOError
            if no files can be found
        """
        if isinstance(pattern, str):
            pattern = (pattern + self._FILEMASK).replace('**', '*')
        else:
            pattern = self._FILEMASK
        if pattern is None:
            const.print_log.warning('_FILEMASK attr. must not be None...'
                                    'using default pattern *.* for file search')
            pattern = '*.*'
        self.logger.info('Fetching data files. This might take a while...')
        files = sorted(glob.glob(os.path.join(self.data_dir,
                                              pattern)))
        if not len(files) > 0:
            all_str = list_to_shortstr(os.listdir(self.data_dir))
            raise DataSourceError('No files could be detected matching file '
                                  'mask {} in dataset {}, files in folder {}:\n'
                                  'Files in folder:{}'.format(pattern,
                                  self.data_id,
                                  self.data_dir,
                                  all_str))
        self.files = files
        return files

    def read_station(self, station_id_filename, **kwargs):
        """Read data from a single station into :class:`UngriddedData`

        Find all files that contain the station ID in their filename and then
        call :func:`read`, providing the reduced filelist as input, in order
        to read all files from this station into data object.

        Parameters
        ----------
        station_id_filename : str
            name of station (MUST be encrypted in filename)
        **kwargs
            additional keyword args passed to :func:`read`
            (e.g. ``vars_to_retrieve``)

        Returns
        -------
        UngriddedData
            loaded data

        Raises
        ------
        IOError
            if no files can be found for this station ID
        """
        files = self.find_in_file_list('*{}*'.format(station_id_filename))
        return self.read(files=files, **kwargs)

    def read_first_file(self, **kwargs):
        """Read first file returned from :func:`get_file_list`

        Note
        ----
        This method may be used for test purposes.

        Parameters
        ----------
        **kwargs
            keyword args passed to :func:`read_file` (e.g. vars_to_retrieve)

        Returns
        -------
        dict-like
            dictionary or similar containing loaded results from first file
        """
        files = self.files
        if len(files) == 0:
            files = self.get_file_list()
        return self.read_file(files[0], **kwargs)

if __name__=="__main__":

    class ReadUngriddedImplementationExample(ReadUngriddedBase):
        _FILEMASK = ".txt"
        DATA_ID = "Blaaa"
        __version__ = "0.01"
        PROVIDES_VARIABLES = ["od550aer"]
        REVISION_FILE = const.REVISION_FILE


        def __init__(self, data_id=None, data_dir=None):
            super(ReadUngriddedImplementationExample, self).__init__(
                data_id, data_dir)

        @property
        def DEFAULT_VARS(self):
            return self.PROVIDES_VARIABLES

        @property
        def SUPPORTED_DATASETS(self):
            return [self.DATA_ID]

        def TS_TYPE(self):
            raise NotImplementedError

        def read(self):
            raise NotImplementedError

        def read_file(self):
            raise NotImplementedError

    c = ReadUngriddedImplementationExample()
    print(c.data_id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Classes and methods to perform high-level colocation.
"""
from datetime import datetime
import glob
import os
from pathlib import Path
import pandas as pd
import traceback

from pyaerocom._lowlevel_helpers import (BrowseDict,
                                         StrWithDefault,
                                         ListOfStrings,
                                         chk_make_subdir)
from pyaerocom import const
from pyaerocom.helpers import (to_pandas_timestamp, to_datestring_YYYYMMDD,
                               get_lowest_resolution, start_stop)
from pyaerocom.colocation import (colocate_gridded_gridded,
                                  colocate_gridded_ungridded,
                                  correct_model_stp_coldata)
from pyaerocom.colocateddata import ColocatedData

from pyaerocom.io import ReadUngridded, ReadGridded, ReadMscwCtm
from pyaerocom.io.helpers import get_all_supported_ids_ungridded
from pyaerocom.exceptions import (ColocationError, ColocationSetupError,
                                  DataCoverageError)

class ColocationSetup(BrowseDict):
    """
    Setup class for high-level model / obs co-location.

    An instance of this setup class can be used to run a colocation analysis
    between a model and an observation network and will create a number of
    :class:`pya.ColocatedData` instances, which can be saved automatically
    as NetCDF files.

    Apart from co-location, this class also handles reading of the input data
    for co-location. Supported co-location options are:

    1. gridded vs. ungridded data
    For instance 3D model data (instance of :class:`GriddedData`) with lat,
    lon and time dimension that is co-located with station based observations
    which are represented in pyaerocom through :class:`UngriddedData` objects.
    The co-location function used is
    :func:`pyaerocom.colocation.colocated_gridded_ungridded`. For this type of
    co-location, the output co-located data object will be 3-dimensional,
    with dimensions `data_source` (index 0: obs, index 1: model), `time` and
    `station_name`.

    2. gridded vs. gridded data
    For instance 3D model data that is co-located with 3D satellite data
    (both instances of :class:`GriddedData`), both objects with lat,
    lon and time dimensions. The co-location function used
    is :func:`pyaerocom.colocation.colocated_gridded_gridded`.
    For this type of co-location, the output co-located data object will be
    4-dimensional,  with dimensions `data_source` (index 0: obs, index 1:
    model), `time` and `latitude` and `longitude`.



    Attributes
    ----------
    model_id : str
        ID of model to be used.
    obs_id : str
        ID of observation network to be used.
    obs_vars : list
        Variables to be analysed (need to be available in input obs dataset).
        Variables that are not available in the model data output will be
        skipped. Alternatively, model variables to be used for a given obs
        variable can also be specified via attributes :attr:`model_use_vars`
        and :attr:`model_add_vars`.
    ts_type : str
        String specifying colocation output frequency.
    start
        Start time of colocation. Input can be integer denoting the year or
        anything that can be converted  into :class:`pandas.Timestamp` using
        :func:`pyaerocom.helpers.to_pandas_timestamp`. If None, than the first
        available date in the model data is used.
    stop
        stop time of colocation. int or anything that can be converted into
        :class:`pandas.Timestamp` using
        :func:`pyaerocom.helpers.to_pandas_timestamp` or None. If None and if
        ``start`` is on resolution of year (e.g. ``start=2010``) then ``stop``
        will be automatically set to the end of that year. Else, it will be
        set to the last available timestamp in the model data.
    filter_name : str
        name of filter to be applied. If None, no filter is used
        (to be precise, if None, then
         :attr:`pyaerocom.const.DEFAULT_REG_FILTER` is used which should
         default to `WORLD-wMOUNTAINS`, that is, no filtering).
    basedir_coldata : str
        Base directory for storing of colocated data files.
    save_coldata : bool
        if True, colocated data objects are saved as NetCDF file.
    obs_name : str, optional
        if provided, this string will be used in colocated data filename to
        specify obsnetwork, else obs_id will be used.
    obs_data_dir : str, optional
        location of obs data. If None, attempt to infer obs location based on
        obs ID.
    obs_use_climatology : bool
        BETA if True, pyaerocom default climatology is computed from observation
        stations (so far only possible for unrgidded / gridded colocation).
    obs_vert_type : str
        AeroCom vertical code encoded in the model filenames (only AeroCom 3
        and later). Specifies which model file should be read in case there are
        multiple options (e.g. surface level data can be read from a
        *Surface*.nc file as well as from a *ModelLevel*.nc file). If input is
        string (e.g. 'Surface'), then the corresponding vertical type code is
        used for reading of all variables that are colocated (i.e. that are
        specified in :attr:`obs_vars`).
    obs_ts_type_read : str or dict, optional
        may be specified to explicitly define the reading frequency of the
        observation data (so far, this does only apply to gridded obsdata such
        as satellites), either as str (same for all obs variables) or variable
        specific as dict. For ungridded reading, the frequency may be specified
        via :attr:`obs_id`, where applicable (e.g. AeronetSunV3Lev2.daily).
        Not to be confused with :attr:`ts_type`, which specifies the
        frequency used for colocation. Can be specified variable specific in
        form of dictionary.
    obs_filters : dict
        filters applied to the observational dataset before co-location.
        In case of gridded / gridded, these are filters that can be passed to
        :func:`pyaerocom.io.ReadGridded.read_var`, for instance, `flex_ts_type`,
        or `constraints`. In case the obsdata is ungridded (gridded / ungridded
        co-locations) these are filters that are handled through keyword
        `filter_post` in :func:`pyaerocom.io.ReadUngridded.read`. These filters
        are applied to the :class:`UngriddedData` objects after reading and
        caching the data, so changing them, will not invalidate the latest
        cache of the :class:`UngriddedData`.
    read_opts_ungridded : dict, optional
        dictionary that specifies reading constraints for ungridded reading,
        and are passed as `**kwargs` to :func:`pyaerocom.io.ReadUngridded.read`.
        Note that - other than for `obs_filters` these filters are applied
        during the reading of the :class:`UngriddedData` objects and specifying
        them will deactivate caching.
    model_name : str, optional
        if provided, this string will be used in colocated data filename to
        specify model, else obs_id will be used.
    model_data_dir : str, optional
        Location of model data. If None, attempt to infer model location based
        on model ID.
    model_read_opts : dict, optional
        options for model reading (passed as keyword args to
        :func:`pyaerocom.io.ReadUngridded.read`).
    model_use_vars : dict, optional
        dictionary that specifies mapping of model variables. Keys are
        observation variables, values are the corresponding model variables
        (e.g. model_use_vars=dict(od550aer='od550csaer')). Example: your
        observation has var *od550aer* but your model model uses a different
        variable name for that variable, say *od550*. Then, you can specify
        this via `model_use_vars = {'od550aer' : 'od550'}`. NOTE: in this case,
        a model variable *od550aer* will be ignored, even if it exists
        (cf :attr:`model_add_vars`).
    model_rename_vars : dict, optional
        rename certain model variables **after** co-location, before storing
        the associated :class:`ColocatedData` object on disk. Keys are model
        variables, values are new names
        (e.g. `model_rename_vars={'od550aer':'MyAOD'}`).
        Note: this does not impact which variables are read from the model.
    model_add_vars : dict, optional
        additional model variables to be processed for one obs variable. E.g.
        `model_add_vars={'od550aer': ['od550so4', 'od550gt1aer']}` would
        co-locate both model SO4 AOD (od550so4) and model coarse mode AOD
        (od550gt1aer) with total AOD (od550aer) from obs (in addition to
        od550aer vs od550aer if applicable).
    model_to_stp : bool
        ALPHA (please do not use): convert model data values to STP conditions
        after co-location. Note: this only works for very particular settings
        at the moment and needs revision, as it relies on access to
        meteorological data.
    model_ts_type_read : str or dict, optional
        may be specified to explicitly define the reading frequency of the
        model data, either as str (same for all obs variables) or variable
        specific as dict. Not to be confused with :attr:`ts_type`, which
        specifies the output frequency of the co-located data.
    model_read_aux : dict, optional
        may be used to specify additional computation methods of variables from
        models. Keys are variables to be computed, values are dictionaries with
        keys `vars_required` (list of required variables for computation of var
        and `fun` (method that takes list of read data objects and computes
        and returns var).
    model_use_climatology : bool
        if True, attempt to use climatological model data field. Note: this
        only works if model data is in AeroCom conventions (climatological
        fields are indicated with 9999 as year in the filename) and if this is
        active, only single year analysis are supported (i.e. provide int to
        :attr:`start` to specify the year and leave :attr:`stop` empty).
    gridded_reader_id : dict
        BETA: dictionary specifying which gridded reader is supposed to be used
        for model (and gridded obs) reading. Note: this is a workaround
        solution and will likely be removed in the future when the gridded
        reading API is more harmonised
        (see https://github.com/metno/pyaerocom/issues/174).
    flex_ts_type : bool
        Bboolean specifying whether reading frequency of gridded data is
        allowed to be flexible. This includes all gridded data, whether it is
        model or gridded observation (e.g. satellites). Defaults to True.
    min_num_obs : dict or int, optional
        time resampling constraints applied, defaults to None, in which case
        no constraints are applied. For instance, say your input is in daily
        resolution and you want output in monthly and you want to make sure to
        have roughly 50% daily coverage for the monthly averages. Then you may
        specify `min_num_obs=15` which will ensure that at least 15 daily
        averages are available to compute a monthly average. However, you may
        also define a hierarchical scheme that first goes from daily to
        weekly and then from weekly to monthly, via a dict. E.g.
        `min_num_obs=dict(monthly=dict(weekly=4), weekly=dict(daily=3))` would
        ensure that each week has at least 3 daily values, as well as that each
        month has at least 4 weekly values.
    resample_how : str or dict, optional
        string specifying how data should be aggregated when resampling in time.
        Default is "mean". Can also be a nested dictionary, e.g.
        `resample_how={'conco3': 'daily': {'hourly' : 'max'}}` would use the
        maximum value to aggregate from hourly to daily for variable conco3,
        rather than the mean.
    obs_remove_outliers : bool
        if True, outliers are removed from obs data before colocation,
        else not. Default is False.
        Custom outlier ranges for each variable can be specified via
        :attr:`obs_outlier_ranges`, and for all other variables, the pyaerocom
        default outlier ranges are used. The latter are specified in
        `variables.ini` file via `minimum` and `maximum` attributes and can
        also be accessed through :attr:`pyaerocom.variable.Variable.minimum`
        and :attr:`pyaerocom.variable.Variable.maximum`, respectively.
    model_remove_outliers : bool
        if True, outliers are removed from model data (normally this should be
        set to False, as the models are supposed to be assessed, including
        outlier cases). Default is False.
        Custom outlier ranges for each variable can be specified via
        :attr:`model_outlier_ranges`, and for all other variables, the pyaerocom
        default outlier ranges are used. The latter are specified in
        `variables.ini` file via `minimum` and `maximum` attributes and can
        also be accessed through :attr:`pyaerocom.variable.Variable.minimum`
        and :attr:`pyaerocom.variable.Variable.maximum`, respectively.
    obs_outlier_ranges : dict, optional
        dictionary specifying outlier ranges for individual obs variables.
        (e.g. dict(od550aer = [-0.05, 10], ang4487aer=[0,4])). Only relevant
        if :attr:`obs_remove_outliers` is True.
    model_outlier_ranges : dict, optional
        like :attr:`obs_outlier_ranges` but for model variables. Only relevant
        if :attr:`model_remove_outliers` is True.
    zeros_to_nan : bool
        If True, zero's in output co-located data object will be converted to
        NaN. Default is False.
    harmonise_units : bool
        if True, units are attempted to be harmonised during co-location
        (note: raises Exception if True and in case units cannot be harmonised).
    regrid_res_deg : int, optional
        resolution in degrees for regridding of model grid (done before
        co-location). Default is None.
    colocate_time : bool
        if True and if obs and model sampling frequency (e.g. daily) are higher
        than output colocation frequency (e.g. monthly), then the datasets are
        first colocated in time (e.g. on a daily basis), before the monthly
        averages are calculated. Default is False.
    reanalyse_existing : bool
        if True, always redo co-location, even if there is already an existing
        co-located NetCDF file (under the output location specified by
        :attr:`basedir_coldata` ) for the given variable combination to be
        co-located. If False and output already exists, then co-location is
        skipped for the associated variable. Default is True.
    raise_exceptions : bool
        if True, Exceptions that may occur for individual variables to be
        processed, are raised, else the analysis is skipped for such cases.
    keep_data : bool
        if True, then all colocated data objects computed when running
        :func:`run` will be stored in :attr:`data`. Defaults to True.
    add_meta : dict
        additional metadata that is supposed to be added to each output
        :class:`ColocatedData` object.
    """
    #: Dictionary specifying alternative vertical types that may be used to
    #: read model data. E.g. consider the variable is  ec550aer,
    #: obs_vert_type='Surface' and obs_vert_type_alt=dict(Surface='ModelLevel').
    #: Now, if a model that is used for the analysis does not contain a data
    #: file for ec550aer at the surface ('*ec550aer*Surface*.nc'), then, the
    #: colocation routine will look for '*ec550aer*ModelLevel*.nc' and if this
    #: exists, it will load it and extract the surface level.
    OBS_VERT_TYPES_ALT = {'Surface'    :   'ModelLevel'}

    #: do not raise Exception if invalid item is attempted to be assigned
    #: (Overwritten from base class)
    CRASH_ON_INVALID = False

    FORBIDDEN_KEYS = [
        'var_outlier_ranges', # deprecated since v0.12.0
        'var_ref_outlier_ranges', # deprecated since v0.12.0
        'remove_outliers' # deprecated since v0.12.0
    ]

    ts_type = StrWithDefault('monthly')
    obs_vars = ListOfStrings()
    def __init__(self, model_id=None, obs_id=None, obs_vars=None,
                 ts_type=None, start=None, stop=None, basedir_coldata=None,
                 save_coldata=False, **kwargs):

        self.model_id = model_id
        self.obs_id = obs_id
        self.obs_vars = obs_vars

        self.ts_type = ts_type
        self.start = start
        self.stop = stop

        #crashes if input filter name is invalid
        self.filter_name = 'WORLD-wMOUNTAINS'

        if basedir_coldata is not None:
            basedir_coldata = self._check_input_basedir_coldata(basedir_coldata)
        else:
            basedir_coldata = const.COLOCATEDDATADIR
        self.basedir_coldata = basedir_coldata
        self.save_coldata = save_coldata

        # END OF ASSIGNMENT OF MOST COMMON PARAMETERS - BELOW ARE FURTHER
        # CONFIG ATTRIBUTES, THAT ARE OPTIONAL AND LESS FREQUENTLY USED

        # Options related to obs reading and processing
        self.obs_name = None
        self.obs_data_dir = None
        self.obs_use_climatology = False

        self._obs_cache_only = False # only relevant if obs is ungridded
        self.obs_vert_type = None
        self.obs_ts_type_read = None
        self.obs_filters = {}

        self.read_opts_ungridded = {}

        # Attributes related to model data
        self.model_name = None
        self.model_data_dir = None

        self.model_read_opts = {}

        self.model_use_vars = {}
        self.model_rename_vars = {}
        self.model_add_vars = {}
        self.model_to_stp = False

        self.model_ts_type_read = None
        self.model_read_aux = {}
        self.model_use_climatology = False

        self.gridded_reader_id = {'model'    : 'ReadGridded',
                                  'obs'      : 'ReadGridded'}

        self.flex_ts_type = True

        # Options related to time resampling
        self.min_num_obs = None
        self.resample_how = 'mean'

        # Options related to outlier removal
        self.obs_remove_outliers = False
        self.model_remove_outliers = False

        # Custom outlier ranges for model and obs
        self.obs_outlier_ranges = {}
        self.model_outlier_ranges = {}

        self.zeros_to_nan = False
        self.harmonise_units = False
        self.regrid_res_deg = None
        self.colocate_time = False

        self.reanalyse_existing = True
        self.raise_exceptions = False
        self.keep_data = True

        self.add_meta = {}
        self.update(**kwargs)

    def _check_input_basedir_coldata(self, basedir_coldata):
        """
        Make sure input basedir_coldata is str and exists

        Parameters
        ----------
        basedir_coldata : str or Path
            basic output directory for colocated data

        Raises
        ------
        ValueError
            If input is invalid.

        Returns
        -------
        str
            valid output directory

        """
        if isinstance(basedir_coldata, Path):
            basedir_coldata = str(basedir_coldata)
        if isinstance(basedir_coldata, str):
            if not os.path.exists(basedir_coldata):
                os.mkdir(basedir_coldata)
            return basedir_coldata
        raise ValueError(
            f'Invalid input for basedir_coldata: {basedir_coldata}'
            )

    def _check_basedir_coldata(self):
        """
        Make sure output directory for colocated data files exists

        Raises
        ------
        FileNotFoundError
            If :attr:`basedir_coldata` does not exist and cannot be created.

        Returns
        -------
        str
            current value of :attr:`basedir_coldata`

        """
        basedir_coldata = self.basedir_coldata
        if basedir_coldata is None:
            basedir_coldata = const.COLOCATEDDATADIR
            if not os.path.exists(basedir_coldata):
                const.print_log.info(f'Creating directory: {basedir_coldata}')
                os.mkdir(basedir_coldata)
        elif isinstance(basedir_coldata, Path):
            basedir_coldata = str(basedir_coldata)
        if isinstance(basedir_coldata, str) and not os.path.exists(basedir_coldata):
            os.mkdir(basedir_coldata)
        if not os.path.exists(basedir_coldata):
            raise FileNotFoundError(
                f'Output directory for colocated data files {basedir_coldata} '
                f'does not exist')
        self.basedir_coldata = basedir_coldata
        return basedir_coldata

    @property
    def basedir_logfiles(self):
        """Base directory for storing logfiles"""
        p = chk_make_subdir(self.basedir_coldata, 'logfiles')
        return p

    def add_glob_meta(self, **kwargs):
        """
        Add global metadata to :attr:`add_meta`

        Parameters
        ----------
        kwargs
            metadata to be added

        Returns
        -------
        None

        """
        self.add_meta.update(**kwargs)

    def __setitem__(self, key, val):
        if key == 'basedir_coldata':
            val = self._check_input_basedir_coldata(val)
        super(ColocationSetup, self).__setitem__(key, val)

    def _period_from_start_stop(self) -> str:
        start, stop = start_stop(self.start, self.stop,
                                 stop_sub_sec=False)
        y0, y1 = start.year, stop.year
        assert y0 <= y1
        if y0 == y1:
            return str(y0)
        else:
            return f'{y0}-{y1}'


class Colocator(ColocationSetup):
    """High level class for running co-location

    Note
    ----
    This object inherits from :class:`ColocationSetup` and is also instantiated
    as such. For setup attributes, please see base class.
    """


    SUPPORTED_GRIDDED_READERS = {
        'ReadGridded' : ReadGridded,
        'ReadMscwCtm' : ReadMscwCtm
    }

    STATUS_CODES = {
        1 : 'SUCCESS',
        2 : 'NOT OK: Missing/invalid model variable',
        3 : 'NOT OK: Missing/invalid obs variable',
        4 : 'NOT OK: Failed to read model variable',
        5 : 'NOT OK: Colocation failed',
        }

    def __init__(self, **kwargs):
        super(Colocator, self).__init__(**kwargs)

        self._log = None
        self.logging = True
        self._loaded_model_data = {}
        self.data = {}

        self._processing_status = []
        self.files_written = []

        self._model_reader = None
        self._obs_reader = None

    @property
    def model_vars(self):
        """
        List of all model variables specified in config

        Note
        ----
        This method does not check if the variables are valid or available.

        Returns
        -------
        list
            list of all model variables specified in this setup.

        """
        ovars = self.obs_vars
        model_vars = []
        for ovar in ovars:
            if ovar in self.model_use_vars:
                model_vars.append(self.model_use_vars[ovar])
            else:
                model_vars.append(ovar)

        for ovar, mvars in self.model_add_vars.items():
            if not ovar in ovars:
                const.print_log.warning(
                    f'Found entry in model_add_vars for obsvar {ovar} which '
                    f'is not specified in attr obs_vars, and will thus be '
                    f'ignored')
            model_vars += mvars
        return model_vars

    @property
    def obs_is_ungridded(self):
        """
        bool: True if obs_id refers to an ungridded observation, else False
        """
        return (True if self.obs_id in get_all_supported_ids_ungridded()
                else False)

    @property
    def model_reader(self):
        """
        Model data reader
        """
        if self._model_reader is not None:
            if self._model_reader.data_id == self.model_id:
                return self._model_reader
            const.print_log.info(
                f'Reloading outdated model reader. ID of current reader: '
                f'{self._model_reader.data_id}. New ID: {self.model_id}'
                )
        self._model_reader = self._instantiate_gridded_reader(what='model')
        self._loaded_model_data = {}
        return self._model_reader

    def _check_data_id_obs_reader(self):
        """
        Check if obs_reader is instantiated with correct obs ID.

        Returns
        -------
        bool
            True if current obs reader can be used for obs reading, else False.
        """
        reader = self._obs_reader
        if reader is None:
            return False
        elif self.obs_is_ungridded and self.obs_id in reader.data_ids:
            return True
        elif self.obs_id==reader.data_id:
            return True

    @property
    def obs_reader(self):
        """
        Observation data reader
        """
        if not self._check_data_id_obs_reader():
            if self.obs_is_ungridded:
                self._obs_reader = ReadUngridded(data_ids=[self.obs_id],
                                                 data_dirs=self.obs_data_dir)
            else:
                self._obs_reader = self._instantiate_gridded_reader(what='obs')
        return self._obs_reader

    @property
    def output_dir(self):
        """
        str: Output directory for colocated data NetCDF files
        """
        self._check_basedir_coldata()
        loc = os.path.join(self.basedir_coldata, self.get_model_name())
        if not os.path.exists(loc):
            const.print_log.info(f'Creating dir {loc}')
            os.mkdir(loc)
        return loc

    @property
    def processing_status(self):
        head = ['Model Var', 'Obs Var', 'Status']
        tab = []
        for mvar, ovar, key in self._processing_status:
            tab.append([mvar, ovar, self.STATUS_CODES[key]])
        return pd.DataFrame(tab, columns=head)

    def get_model_name(self):
        """
        Get name of model

        Note
        ----
        Not to be confused with :attr:`model_id` which is always the database
        ID of the model, while model_name can differ from that and is used for
        output files, etc.

        Raises
        ------
        AttributeError
            If neither model_id or model_name are set

        Returns
        -------
        str
            preferably :attr:`model_name`, else :attr:`model_id`

        """
        if self.model_name is None:
            if self.model_id is None:
                raise AttributeError('Neither model_name nor model_id are set')
            return self.model_id
        return self.model_name

    def get_obs_name(self):
        """
        Get name of obsdata source

        Note
        ----
        Not to be confused with :attr:`obs_id` which is always the database
        ID of the observation dataset, while obs_name can differ from that
        and is used for output files, etc.

        Raises
        ------
        AttributeError
            If neither obs_id or obs_name are set

        Returns
        -------
        str
            preferably :attr:`obs_name`, else :attr:`obs_id`

        """
        if self.obs_name is None:
            if self.obs_id is None:
                raise AttributeError('Neither obs_name nor obs_id are set')
            return self.obs_id
        return self.obs_name

    def get_model_data(self, model_var):
        if model_var in self._loaded_model_data:
            mdata = self._loaded_model_data[model_var]
            if mdata.data_id == self.model_id:
                return mdata
        mdata = self._read_gridded(var_name=model_var, is_model=True)
        self._loaded_model_data[model_var] = mdata
        return mdata

    def get_obs_data(self, obs_var):
        if self.obs_is_ungridded:
            return self._read_ungridded(obs_var)
        else:
            return self._read_gridded(obs_var, is_model=False)

    def get_start_str(self):
        if self.start is None:
            raise AttributeError('start time is not set')
        return to_datestring_YYYYMMDD(to_pandas_timestamp(self.start))

    def get_stop_str(self):
        if self.stop is None:
            raise AttributeError('stop time is not set')
        return to_datestring_YYYYMMDD(to_pandas_timestamp(self.stop))

    def prepare_run(self, var_list: list = None) -> dict:
        """
        Prepare colocation run for current setup.

        Parameters
        ----------
        var_name : str, optional
            Variable name that is supposed to be analysed. The default is None,
            in which case all defined variables are attempted to be colocated.

        Raises
        ------
        AttributeError
            If no observation variables are defined (:attr:`obs_vars` empty).

        Returns
        -------
        vars_to_process : dict
            Mapping of variables to be processed, keys are model vars, values
            are obs vars.

        """
        try:
            self._init_log()
        except Exception:
            const.print_log.warning('Deactivating logging in Colocator')
            self.logging = False

        if isinstance(self.obs_vars, str):
            self.obs_vars = [self.obs_vars]
        elif not isinstance(self.obs_vars, list):
            raise AttributeError(
                'obs_vars not defined or invalid, need list with strings...'
                )
        self._check_obs_vars_available()
        self._check_obs_filters()
        self._check_model_add_vars()
        self._check_set_start_stop()

        vars_to_process = self._find_var_matches()
        if var_list is not None:
           vars_to_process = self._filter_var_matches_varlist(vars_to_process,
                                                              var_list)

        (vars_to_process,
         ts_types) = self._check_load_model_data(vars_to_process)

        if self.save_coldata and not self.reanalyse_existing:
            vars_to_process = self._filter_var_matches_files_not_exist(
                vars_to_process, ts_types)
        return vars_to_process

    def run(self, var_list: list = None, **opts):
        """Perform colocation for current setup

        See also :func:`prepare_run`.

        Parameters
        ----------
        var_list : list, optional
            list of variables supposed to be analysed. The default is None,
            in which case all defined variables are attempted to be colocated.
        **opts
            keyword args that may be specified to change the current setup
            before colocation

        Returns
        -------
        dict
            nested dictionary, where keys are model variables, values are
            dictionaries comprising key / value pairs of obs variables and
            associated instances of :class:`ColocatedData`.
        """
        self.update(**opts)
        data_out = {}
        # ToDo: see if the following could be solved via custom context manager
        try:
            vars_to_process = self.prepare_run(var_list)
        except Exception:
            if self.raise_exceptions:
                self._print_processing_status()
                self._write_log('ABORTED: raise_exceptions is True\n')
                self._close_log()
                raise
            vars_to_process = {}
        self._print_coloc_info(vars_to_process)
        for mod_var, obs_var in vars_to_process.items():
            try:
                coldata = self._run_helper(mod_var, obs_var)
                if not mod_var in data_out:
                    data_out[mod_var] = {}
                data_out[mod_var][obs_var] = coldata
                self._processing_status.append([mod_var, obs_var, 1])
            except Exception:
                msg = ('Failed to perform analysis: {}\n'
                       .format(traceback.format_exc()))
                const.print_log.warning(msg)
                self._processing_status.append([mod_var, obs_var, 5])
                self._write_log(msg)
                if self.raise_exceptions:
                    self._print_processing_status()
                    self._write_log('ABORTED: raise_exceptions is True\n')
                    self._close_log()
                    raise ColocationError(traceback.format_exc())
        self._write_log('Colocation finished')
        self._close_log()
        self._print_processing_status()
        if self.keep_data:
            self.data = data_out
        return data_out

    def get_nc_files_in_coldatadir(self):
        """
        Get list of NetCDF files in colocated data directory

        Returns
        -------
        list
            list of NetCDF file paths found

        """
        mask = f'{self.output_dir}/*.nc'
        return glob.glob(mask)

    def get_available_coldata_files(self, var_list : list = None) -> list:
        self._check_set_start_stop()

        def check_meta_match(meta, **kwargs):
            for key, val in kwargs.items():
                if not meta[key] == val:
                    return False
            return True

        mname = self.get_model_name()
        oname = self.get_obs_name()
        model_vars = self.model_vars
        obs_vars = self.obs_vars
        start, stop = self.get_start_str(), self.get_stop_str()
        valid = []
        all_files = self.get_nc_files_in_coldatadir()
        for file in all_files:
            try:
                meta = ColocatedData.get_meta_from_filename(file)
            except Exception:
                continue
            candidate =  check_meta_match(meta, model_name=mname, obs_name=oname,
                                          start=start, stop=stop)
            if (candidate and meta['model_var'] in model_vars and
                meta['obs_var'] in obs_vars):
                if var_list is None:
                    ok = True
                else:
                    ok = False
                    for var in var_list:
                        if meta['model_var'] == var or meta['obs_var'] == var:
                            ok = True
                            break
                if ok:
                    valid.append(file)

        return valid

    def _filter_var_matches_varlist(self, vars_to_process, var_list) -> dict:
        _vars_to_process = {}
        if isinstance(var_list, str):
            var_list = [var_list]
        for var_name in var_list:
            _subset = self._filter_var_matches_var_name(
                vars_to_process, var_name)
            _vars_to_process.update(**_subset)
        return _vars_to_process

    def _read_ungridded(self, var_name):
        """Helper to read UngriddedData

        Note
        ----
        reading is restricted to single variable UngriddedData objects here,
        due to multiple possibilities for variable specific obs filter
        definitions and because the colocation is done sequentially for each
        variable.

        Parameters
        ----------
        vars_to_read : str or list, optional
            variables that should be read from obs-network (:attr:`obs_id`)

        Returns
        -------
        UngriddedData
            loaded data object

        """
        obs_reader = self.obs_reader
        obs_filters_post = self._eval_obs_filters(var_name)

        obs_data = obs_reader.read(
            data_ids=[self.obs_id],
            vars_to_retrieve=var_name,
            only_cached=self._obs_cache_only,
            filter_post=obs_filters_post,
            **self.read_opts_ungridded)

        if self.obs_remove_outliers:
            oor = self.obs_outlier_ranges
            if var_name in oor:
                low, high = oor[var_name]
            else:
                var_info = const.VARS[var_name]
                low, high = var_info.minimum, var_info.maximum
            obs_data.remove_outliers(var_name,
                                     low=low,high=high,
                                     inplace=True,
                                     move_to_trash=False)
        return obs_data

    def _check_obs_filters(self):
        obs_vars = self.obs_vars
        if any([x in self.obs_filters for x in obs_vars]):
            # variable specific obs_filters
            for ovar in obs_vars:
                if not ovar in self.obs_filters:
                    self.obs_filters[ovar] = {}

    def _check_load_model_data(self, var_matches):
        """
        Try to preload modeldata for input variable matches

        Note
        ----
        This will load all model fields for each obs variable in lazy mode, so
        should not require much storage. T

        Parameters
        ----------
        var_matches : dict
            dictionary specifying model / obs var pairs for colocation.

        Raises
        ------
        ColocationError
            If :attr:`raise_exceptions` is True and if one of the input
            model variables cannot be loaded.

        Returns
        -------
        filtered : dict
            `var_matches` filtered by entries for which model data could be
            successfully read.
        ts_types : dict
            data frequencies for each model variable that could be read. Those
            depend also on settings for :attr:`flex_ts_type`

        """
        filtered, ts_types = {}, {}
        for mvar, ovar in var_matches.items():
            try:
                mdata = self.get_model_data(mvar)
                filtered[mvar] = ovar
                ts_types[mvar] = mdata.ts_type
            except Exception as e:
                msg = (
                    f'Failed to load model data: {self.model_id} ({mvar}). '
                    f'Reason {e}')
                const.print_log.warning(msg)
                self._write_log(msg + '\n')
                self._processing_status.append([mvar, ovar, 4])
                if self.raise_exceptions:
                    raise ColocationError(msg)
        return filtered, ts_types

    def _filter_var_matches_files_not_exist(self, var_matches, ts_types):
        filtered = {}
        for mvar, ovar in var_matches.items():
            ts_type = ts_types[mvar]
            fname = self._coldata_savename(ovar, mvar, ts_type)
            fp = os.path.join(self.output_dir, fname)
            if not os.path.exists(fp):
                filtered[mvar] = ovar
        return filtered

    def _check_model_add_vars(self):
        for ovar, mvars in self.model_add_vars.items():
            if not isinstance(mvars, list):
                raise ValueError('Values of model_add_vars need to be list')
            elif not all([isinstance(x, str) for x in mvars]):
                raise ValueError(
                    'Values of model_add_vars need to be list of strings')

    def _instantiate_gridded_reader(self, what):
        """
        Create reader for model or observational gridded data.

        Parameters
        ----------
        what : str
            Type of reader. ("model" or "obs")

        Returns
        -------
        Instance of reader class defined in self.SUPPORTED_GRIDDED_READERS
        """
        if what == 'model':
            data_id = self.model_id
            data_dir = self.model_data_dir
        else:
            data_id = self.obs_id
            data_dir = self.obs_data_dir
        reader_class = self._get_gridded_reader_class(what=what)
        reader = reader_class(data_id=data_id, data_dir=data_dir)
        if isinstance(reader, ReadMscwCtm) and hasattr(self, 'filepath'):
            reader.filepath = self.filepath
        return reader

    def _get_gridded_reader_class(self, what):
        """Returns the class of the reader for gridded data."""
        try:
            reader = self.SUPPORTED_GRIDDED_READERS[self.gridded_reader_id[what]]
        except KeyError as e:
            raise NotImplementedError('Reader {} is not supported: {}'.
                                      format(self.gridded_reader_id[what], e))
        return reader

    def _check_add_model_read_aux(self, model_var):

        if not model_var in self.model_read_aux:
            return False
        info = self.model_read_aux[model_var]
        if not isinstance(info, dict):
            raise ValueError('Invalid value for model_read_aux of variable '
                             '{}. Need dictionary, got {}'
                             .format(model_var, info))
        elif not all([x in info for x in ['vars_required', 'fun']]):
            raise ValueError('Invalid value for model_read_aux dict of variable '
                             '{}. Require keys vars_required and fun in dict, '
                             'got {}'.format(model_var, info))
        try:
            self.model_reader.add_aux_compute(var_name=model_var, **info)
        except DataCoverageError:
            return False
        return True

    def _check_obs_vars_available(self):
        if self.obs_vars == []:
            raise ColocationSetupError('no observation variables specified...')
        oreader = self.obs_reader
        if self.obs_is_ungridded:
            avail = oreader.get_vars_supported(self.obs_id, self.obs_vars)
        else:
            avail = []
            for ovar in self.obs_vars:
                if oreader.has_var(ovar):
                    avail.append(ovar)

        if len(self.obs_vars) > len(avail):
            for ovar in self.obs_vars:
                if not ovar in avail:
                    const.print_log.warning(
                        f'Obs variable {ovar} is not available in {self.obs_id} '
                        f'and will be ignored'
                        )
                    self._processing_status.append([None, ovar, 3])

            if self.raise_exceptions:
                invalid = [var for var in self.obs_vars if not var in avail]
                invalid = '; '.join(invalid)
                raise DataCoverageError(
                    f'Invalid obs var(s) for {self.obs_id}: {invalid}')

            self.obs_vars = avail

    def _print_processing_status(self):

        mname = self.get_model_name()
        oname = self.get_obs_name()
        const.print_log.info(
            f'Colocation processing status for {mname} vs. {oname}')
        const.print_log.info(self.processing_status)

    def _filter_var_matches_var_name(self, var_matches, var_name):
        filtered = {}
        for mvar, ovar in var_matches.items():
            if mvar in var_name or ovar in var_name:
                filtered[mvar] = ovar
        if len(filtered) == 0:
            raise DataCoverageError(var_name)
        return filtered

    def _find_var_matches(self):
        """Find variable matches in model data for input obs variables"""

        # dictionary that will map model variables (keys) with observation
        # variables (values)
        var_matches = {}

        all_ok = True
        muv = self.model_use_vars
        modreader = self.model_reader
        for ovar in self.obs_vars:
            if ovar in muv:
                mvar = muv[ovar]
            else:
                mvar = ovar
            self._check_add_model_read_aux(mvar)
            if modreader.has_var(mvar):
                var_matches[mvar] = ovar
            else:
                self._processing_status.append([mvar, ovar, 2])
                all_ok = False

            if ovar in self.model_add_vars: #observation variable
                addvars = self.model_add_vars[ovar]
                for addvar in addvars:
                    self._check_add_model_read_aux(addvar)
                    if modreader.has_var(addvar):
                        var_matches[addvar] = ovar
                    else:
                        self._processing_status.append([addvar, ovar, 2])
                        all_ok = False

        if not all_ok and self.raise_exceptions:
            raise DataCoverageError(
                'Some model variables are not available')

        return var_matches

    def _get_ts_type_read(self, var_name, is_model):
        """
        Get *desired* reading frequency for gridded reading

        Parameters
        ----------
        var_name : str
            Name of variable to be read.
        is_model : bool
            True if reading refers to model reading, else False (e.g. gridded
            satellite obs).

        Raises
        ------
        ValueError


        Returns
        -------
        str or None
            frequency to be read.

        """
        tst = self.ts_type # default
        if is_model and self.model_ts_type_read is not None:
            tst = self.model_ts_type_read
            if tst == '':
                tst = self.ts_type
        elif not is_model and self.obs_ts_type_read is not None:
            tst = self.obs_ts_type_read
        if isinstance(tst, dict):
            if var_name in tst:
                tst = tst[var_name]
            else:
                tst = self.ts_type
        return tst

    def _read_gridded(self, var_name, is_model):

        start, stop = self.start, self.stop
        ts_type_read = self._get_ts_type_read(var_name, is_model)
        kwargs = {}
        if is_model:
            reader = self.model_reader
            vert_which = self.obs_vert_type
            if self.model_use_climatology:
                # overwrite start and stop to read climatology file for model
                start, stop = 9999, None
            if var_name in self.model_read_opts:
                kwargs.update(self.model_read_opts[var_name])
        else:
            reader = self.obs_reader
            vert_which = None
            ts_type_read = self.obs_ts_type_read
            kwargs.update(self._eval_obs_filters(var_name))

        try:
            data = reader.read_var(var_name,
                                   start=start,
                                   stop=stop,
                                   ts_type=ts_type_read,
                                   vert_which=vert_which,
                                   flex_ts_type=self.flex_ts_type,
                                   **kwargs)
        except DataCoverageError:
            vert_which_alt = self._try_get_vert_which_alt(is_model, var_name)
            data = reader.read_var(var_name,
                                   start=start,
                                   stop=stop,
                                   ts_type=ts_type_read,
                                   flex_ts_type=self.flex_ts_type,
                                   vert_which=vert_which_alt)

        data = self._check_remove_outliers_gridded(data, var_name, is_model)
        return data

    def _try_get_vert_which_alt(self, is_model, var_name):
        if is_model:
            if self.obs_vert_type in self.OBS_VERT_TYPES_ALT:
                return self.OBS_VERT_TYPES_ALT[self.obs_vert_type]
        raise DataCoverageError(f'No alternative vert type found for {var_name}')

    def _check_remove_outliers_gridded(self, data, var_name, is_model):
        if is_model:
            rm_outliers = self.model_remove_outliers
            outlier_ranges = self.model_outlier_ranges
        else:
            rm_outliers = self.obs_remove_outliers
            outlier_ranges = self.obs_outlier_ranges

        if len(outlier_ranges) > 0 and not rm_outliers:
            const.print_log.warning(
                f'WARNING: Found definition of outlier ranges for {var_name} '
                f'({data.data_id}) but outlier removal is deactivated. Consider '
                f'checking your setup (note: model or obs outlier removal can be '
                f'activated via attrs. model_remove_outliers and remove_outliers, '
                f'respectively')

        if rm_outliers:
            if var_name in outlier_ranges:
                low, high = outlier_ranges[var_name]
            else:
                var_info = const.VARS[var_name]
                low, high = var_info.minimum, var_info.maximum
            data.check_unit()
            data.remove_outliers(low, high, inplace=True)
        return data

    def _eval_obs_filters(self, var_name):
        obs_filters = self['obs_filters']
        if var_name in obs_filters:
            obs_filters = obs_filters[var_name]
        remaining = {}
        if not isinstance(obs_filters, dict):
            raise AttributeError('Detected obs_filters attribute in '
                                 'Colocator class, which is not a '
                                 'dictionary: {}'.format(obs_filters))
        for key, val in obs_filters.items():
            # keep ts_type filter in remaining (added on 17.2.21, 0.10.0 -> 0.10.1)
            if key in self and not key == 'ts_type': # can be handled
                if isinstance(self[key], dict) and isinstance(val, dict):
                    self[key].update(val)
                else:
                    self[key] = val
            else:
                remaining[key] = val
        return remaining

    def _save_coldata(self, coldata):
        """Helper for saving colocateddata"""
        obs_var, mod_var = coldata.metadata['var_name']
        if mod_var in self.model_rename_vars:
            mvar = self.model_rename_vars[mod_var]
            const.print_log.info(
                f'Renaming model variable from {mod_var} to {mvar} in '
                f'ColocatedData before saving to NetCDF.')
            coldata.rename_variable(mod_var, mvar,
                                    self.model_id)
        else:
            mvar = mod_var
        savename = self._coldata_savename(obs_var, mvar, coldata.ts_type)
        fp = coldata.to_netcdf(self.output_dir, savename=savename)
        self.files_written.append(fp)
        msg = f'WRITE: {fp}\n'
        self._write_log(msg)
        const.print_log.info(msg)

    def _eval_resample_how(self, model_var, obs_var):
        rshow = self.resample_how
        if not isinstance(rshow, dict):
            return rshow

        if obs_var in rshow:
            return rshow[obs_var]
        elif model_var in rshow:
            return rshow[model_var]
        else:
            return None

    def _infer_start_stop_yr_from_model_reader(self):
        """
        Infer start / stop year for colocation from gridded model reader

        Sets :attr:`start` and :attr:`stop`

        """
        # get sorted list of years available in model data (files with year
        # 9999 denote climatological data)
        yrs_avail = self.model_reader.years_avail
        if self.model_use_climatology:
            if not 9999 in yrs_avail:
                raise DataCoverageError('No climatology files available')
            first, last = 9999, None
        else:
            if 9999 in yrs_avail:
                yrs_avail = [x for x in yrs_avail if not x==9999]
            first, last = yrs_avail[0], yrs_avail[-1]
            if first == last:
                last = None
        self.start = first
        self.stop = last

    def _check_set_start_stop(self):
        if self.start is None:
            self._infer_start_stop_yr_from_model_reader()
        if self.model_use_climatology:
            if self.stop is not None or not isinstance(self.start, int):
                raise ColocationSetupError(
                'Conflict: only single year analyses are support for model '
                'climatology fields, please specify "start" as integer '
                'denoting the year, and set "stop"=None')
        self.start, self.stop = start_stop(self.start, self.stop)


    def _coldata_savename(self, obs_var, mod_var, ts_type):
        """Get filename of colocated data file for saving
        """
        name = ColocatedData._aerocom_savename(
            obs_var=obs_var, obs_id=self.get_obs_name(),
            mod_var=mod_var, mod_id=self.get_model_name(),
            start_str=self.get_start_str(),
            stop_str=self.get_stop_str(),
            ts_type=ts_type,
            filter_name=self.filter_name
            )
        return f'{name}.nc'


    def _get_colocation_ts_type(self, model_ts_type, obs_ts_type=None):
        chk = [self.ts_type, model_ts_type]
        if obs_ts_type is not None:
            chk.append(obs_ts_type)
        return get_lowest_resolution(*chk)

    @property
    def _colocation_func(self):
        """
        Function used for colocation

        Returns
        -------
        callable
            function the performs co-location operation

        """
        if self.obs_is_ungridded:
            return colocate_gridded_ungridded
        else:
            return colocate_gridded_gridded

    def _prepare_colocation_args(self, model_var, obs_var):

        model_data = self.get_model_data(model_var)
        obs_data = self.get_obs_data(obs_var)
        rshow = self._eval_resample_how(model_var, obs_var)

        if self.model_use_climatology:
            baseyr = self.start.year
        else:
            baseyr = None
        # input args shared between all colocation functions
        args = dict(
            data=model_data,
            data_ref=obs_data,
            start=self.start,
            stop=self.stop,
            filter_name=self.filter_name,
            regrid_res_deg=self.regrid_res_deg,
            harmonise_units=self.harmonise_units,
            update_baseyear_gridded=baseyr,
            min_num_obs=self.min_num_obs,
            colocate_time=self.colocate_time,
            resample_how=rshow
            )
        if self.obs_is_ungridded:
            ts_type = self._get_colocation_ts_type(model_data.ts_type)
            args.update(ts_type=ts_type, var_ref=obs_var,
                        use_climatology_ref=self.obs_use_climatology)
        else:
            ts_type = self._get_colocation_ts_type(model_data.ts_type,
                                                   obs_data.ts_type)
            args.update(ts_type=ts_type)
        return args

    def _check_dimensionality(self, args):
        mdata = args['data']
        odata = args['data_ref']
        from pyaerocom.griddeddata import GriddedData
        from pyaerocom.exceptions import DataDimensionError
        if mdata.ndim==4 and self.obs_vert_type == 'Surface':
            mdata = mdata.extract_surface_level()
            args['data'] = mdata
        elif mdata.ndim > 3:
            raise DataDimensionError(f'cannot co-locate model data with more '
                                     f'than 3 dimensions: {mdata}')

        if isinstance(odata, GriddedData):
            if odata.ndim == 4 and self.obs_vert_type == 'Surface':
                odata = odata.extract_surface_level()
                args['data_ref'] = odata
            elif odata.ndim > 3:
                raise DataDimensionError(
                    f'cannot co-locate model data with more '
                    f'than 3 dimensions: {odata}')
        return args

    def _run_helper(self, model_var, obs_var):
        const.print_log.info(
            f'Running {self.model_id} ({model_var}) vs. '
            f'{self.obs_id} ({obs_var})'
            )
        args = self._prepare_colocation_args(model_var, obs_var)
        args = self._check_dimensionality(args)
        coldata = self._colocation_func(**args)

        coldata.data.attrs['model_name'] = self.get_model_name()
        coldata.data.attrs['obs_name'] = self.get_obs_name()
        coldata.data.attrs['vert_code'] = self.obs_vert_type
        coldata.data.attrs.update(**self.add_meta)

        if self.zeros_to_nan:
            coldata = coldata.set_zeros_nan()
        if self.model_to_stp:
            coldata = correct_model_stp_coldata(coldata)
        if self.save_coldata:
            self._save_coldata(coldata)

        return coldata

    def _print_coloc_info(self, var_matches):
        if not var_matches:
            const.print_log.info('Nothing to colocate')
            return
        const.print_log.info(
            'The following variable combinations will be colocated\n'
            'MODEL-VAR\tOBS-VAR')

        for key, val in var_matches.items():
            const.print_log.info('{}\t{}'.format(key, val))

    def _init_log(self):
        logdir = chk_make_subdir(self.basedir_logfiles,
                                 self.get_model_name())
        oname = self.get_obs_name()
        datestr = datetime.today().strftime('%Y%m%d')
        datetimestr = datetime.today().strftime('%d-%m-%Y %H:%M')

        fname = (f'{oname}_{datestr}.log')
        logfile = os.path.join(logdir, fname)
        self._log = log = open(logfile, 'a+')
        log.write('\n------------------ NEW ----------------\n')
        log.write(f'Timestamp: {datetimestr}\n\n')
        log.write('Analysis configuration\n')
        ignore = ['_log', 'logging', 'data', '_model_reader', '_obs_reader']
        for key, val in self.items():
            if key in ignore:
                continue
            log.write(f'{key}: {val}\n')

    def _write_log(self, msg):
        if self.logging:
            self._log.write(msg)

    def _close_log(self):
        if self._log is not None:
            self._log.close()
            self._log = None




if __name__ == '__main__':
    import matplotlib.pyplot as plt
    plt.close('all')
    col = ColocationSetup()
    col.update(bla=42)



import os
import pandas as pd
from tqdm import tqdm
import numpy as np
from pyaerocom import const
from pyaerocom.exceptions import DataRetrievalError
from pyaerocom.helpers import varlist_aerocom
from pyaerocom.aux_var_helpers import concx_to_vmrx
from pyaerocom.molmasses import get_molmass
from pyaerocom.ungriddeddata import UngriddedData
from geonum import atmosphere as atm
from pyaerocom.io import ReadUngriddedBase

P_STD = atm.p0 # standard atmosphere pressure
T_STD = atm.T0_STD # standard atmosphere temperature

def _conc_to_vmr_marcopolo_stats(data, to_var, from_var,
                                 p_pascal=None, T_kelvin=None,
                                 mmol_air=None):
    if p_pascal is None:
        p_pascal = P_STD
    if T_kelvin is None:
        T_kelvin =  T_STD
    if mmol_air is None:
        mmol_air = get_molmass('air_dry')

    for stat in data:
        if not from_var in stat:
            continue

        concdata = stat[from_var]

        mmol_var = get_molmass(to_var)
        from_unit = stat['var_info'][from_var]['units']
        to_unit = const.VARS[to_var].units
        vmrvals = concx_to_vmrx(concdata,
                             p_pascal=p_pascal,
                             T_kelvin=T_kelvin,
                             mmol_var= mmol_var,
                             mmol_air=mmol_air,
                             conc_unit=from_unit,
                             to_unit=to_unit)
        stat[to_var] = vmrvals
        vi = {}
        vi.update(stat['var_info'][from_var])
        vi['computed'] = True
        vi['units'] = to_unit
        vi['units_info'] = (
            f'The original data is provided as mass conc. in units of ug m-3 and '
            f'was converted to ppb assuming a standard atmosphere '
            f'(p={p_pascal/100}hPa, T={T_kelvin}K) assuming dry molar mass of air '
            f'M_Air_dry={mmol_air}g/mol and total molecular mass of '
            f'{from_var}={mmol_var}g/mol'
            )
        stat['var_info'][to_var] = vi

    return data

class ReadMarcoPolo(ReadUngriddedBase):
    """
    Reading routine for chinese MarcoPolo observations
    """

    #: Data type of files
    _FILETYPE = '.csv'

    #: Filemask for glob data file search
    _FILEMASK = f'*{_FILETYPE}'

    #: Column delimiter
    FILE_COL_DELIM = ','

    #: Version log of this class (for caching)
    __version__ = '0.02'

    #: Default data ID
    DATA_ID = 'MarcoPolo'

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [DATA_ID]

    #: Indices of columns in data files
    FILE_COLS = {
        'station_id'    : 0,
        'datetime'      : 1,
        'var_name'      : 2,
        'value'         : 3
        }
    #: Mapping of columns in station metadata file to pyaerocom standard
    STATION_META_MAP = {
        'stationId'     : 'station_id',
        'stationName'   : 'station_name',
        'cityId'        : 'city_id',
        'latitude'      : 'latitude',
        'longitude'     : 'longitude',
        'cityName'      : 'city'
        }

    #: conversion functions for metadata dtypes
    STATION_META_DTYPES = {
        'station_id'     : str,
        'station_name'   : str,
        'city_id'        : str,
        'latitude'      : float,
        'longitude'     : float,
        'city'      : str
        }

    #: Variable names in data files
    VAR_MAP = {
        'concco': 'co',
        'concso2': 'so2',
        'concno2': 'no2',
        'conco3': 'o3',
        'concpm10': 'pm10',
        'concpm25': 'pm2_5'
        }

    #: Variable units (not provided in data files)
    VAR_UNITS = {
        'concco'   : 'unknown',
        'concso2'  : 'unknown',
        'concno2'  : 'ug m-3', # mail from Henk Eskes, 1 Feb. 2021
        'conco3'   : 'ug m-3', # mail from Henk Eskes, 1 Feb. 2021
        'concpm10' : 'ug m-3', # mail from Henk Eskes, 1 Feb. 2021
        'concpm25' : 'ug m-3' # mail from Henk Eskes, 1 Feb. 2021
        }

    #: Variables that are computed (cannot be read directly)
    AUX_REQUIRES = {
        'vmro3'  : ['conco3'],
        'vmrno2' : ['concno2']
        }

    #: functions used to convert variables that are computed
    AUX_FUNS = {
        'vmro3'  : _conc_to_vmr_marcopolo_stats,
        'vmrno2' : _conc_to_vmr_marcopolo_stats
        }


    PROVIDES_VARIABLES = list(VAR_MAP.keys())
    DEFAULT_VARS = ['concpm10', 'concpm25', 'concno2', 'conco3']

    TS_TYPE = 'hourly'

    STAT_METADATA_FILENAME = 'station_info.xlsx'

    def __init__(self, data_id=None, data_dir=None):
        super(ReadMarcoPolo, self).__init__(data_id=data_id,
                                            data_dir=data_dir)

        try:
            import openpyxl
        except ImportError:
            raise ImportError(
                'ReadMarcoPolo class needs package openpyxl '
                'which is not part of the pyaerocom standard installation. '
                'Please install openpyxl via conda or pip.')

    def _read_metadata_file(self):
        fn = os.path.join(self.data_dir, self.STAT_METADATA_FILENAME)
        cfg = pd.read_excel(fn,engine='openpyxl')
        return cfg

    def _init_station_metadata(self):
        """
        Initiate metadata for all stations

        Returns
        -------
        dict
            dictionary with metadata dictionaries for all stations

        """

        cfg = self._read_metadata_file()

        meta_map = self.STATION_META_MAP
        dtypes = self.STATION_META_DTYPES

        cols = list(cfg.columns.values)
        col_idx = {}
        for from_meta, to_meta in meta_map.items():
            col_idx[to_meta] = cols.index(from_meta)

        arr = cfg.values

        stats = {}
        for row in arr:
            stat = {}
            for meta_key, col_num in col_idx.items():
                stat[meta_key] = dtypes[meta_key](row[col_num])
            stat['data_id'] = self.data_id
            stat['ts_type'] = self.TS_TYPE
            stats[stat['station_id']] = stat

        return stats

    def _read_file(self, file):
        df = pd.read_csv(file, sep=self.FILE_COL_DELIM)
        return df

    def _make_station_data(self, var, var_stats_unique, stat_ids, var_stats,
                           var_dtime, var_values, stat_meta):

        units = self.VAR_UNITS
        stats = []
        for stat_id in var_stats_unique:
            if not stat_id in stat_ids:
                continue

            statmask = var_stats == stat_id

            if statmask.sum() == 0:
                continue

            timestamps = var_dtime[statmask]
            vals = var_values[statmask]



            stat = {}
            stat.update(stat_meta[stat_id])

            stat['dtime'] = timestamps
            stat['timezone'] ='unknown'
            stat[var] = vals
            unit = units[var]
            stat['var_info'] = {}
            stat['var_info'][var] = dict(units=unit)
            stats.append(stat)
        return stats

    def _read_files(self, files, vars_to_retrieve):

        filecols = self.FILE_COLS

        stat_meta = self._init_station_metadata()

        # get all station IDs found in the metadata file
        stat_ids = list(stat_meta.keys())

        arrs = []

        varcol = filecols['var_name']
        for i in tqdm(range(len(files))):
            filedata = self._read_file(files[i])
            arr = filedata.values

            for i, var in enumerate(vars_to_retrieve):
                filevar = self.VAR_MAP[var]
                if i == 0:
                    mask = arr[:, varcol] == filevar
                else:
                    mask = np.logical_or(mask, arr[:, varcol] == filevar)
            matches = mask.sum()
            if matches:
                vardata = arr[mask]
                arrs.append(vardata)
        if len(arrs) == 0:
            raise DataRetrievalError(
                'None of the input variables could be found in input list')
        data = np.concatenate(arrs)

        dtime = data[:, filecols['datetime']].astype('datetime64[s]')

        varis = data[:, varcol].astype(str)
        values = data[:, filecols['value']].astype(np.float64)
        statids = data[:, filecols['station_id']].astype(str)

        stats = []
        for var in vars_to_retrieve:
            var_in_file = self.VAR_MAP[var]

            varmask = varis == var_in_file

            var_values = values[varmask]
            var_stats = statids[varmask]
            var_dtime = dtime[varmask]

            var_stats_unique = np.unique(var_stats)

            var_stats = self._make_station_data(var, var_stats_unique,
                                                stat_ids, var_stats,
                                                var_dtime, var_values,
                                                stat_meta)
            stats.extend(var_stats)

        return stats

    def read_file(self):
        """
        This method is not implemented (but needs to be declared for template)

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError('Not needed for these data since the format '
                                  'is unsuitable...')

    def compute_additional_vars(self, statlist_from_file, vars_to_compute):

        for var in vars_to_compute:
            fun = self.AUX_FUNS[var]
            requires = self.AUX_REQUIRES[var]
            # this will add the variable data to each station data in
            # statlist_from_file
            statlist_from_file = fun(statlist_from_file, var, *requires)


        return statlist_from_file

    def read(self, vars_to_retrieve=None, first_file=None, last_file=None):
        """
        Read variable data

        Parameters
        ----------
        vars_to_retrieve : str or list, optional
            List of variables to be retrieved. The default is None.
        first_file : int, optional
            Index of first file to be read. The default is None, in which case
            index 0 in file list is used.
        last_file : int, optional
            Index of last file to be read. The default is None, in which case
            last index in file list is used.

        Returns
        -------
        data : UngriddedData
            loaded data object.

        """
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]

        # make sure to use AeroCom variable names in output data
        vars_to_retrieve = varlist_aerocom(vars_to_retrieve)

        vars_to_read, vars_to_compute = self.check_vars_to_retrieve(vars_to_retrieve)

        files = self.get_file_list()
        if first_file is None:
            first_file = 0
        if last_file is None:
            last_file = len(files)
        files = files[first_file:last_file]

        stats = self._read_files(files, vars_to_read)

        stats = self.compute_additional_vars(stats, vars_to_compute)

        data = UngriddedData.from_station_data(stats)

        return data

if __name__ == '__main__':
    path_data = '/home/jonasg/MyPyaerocom/data/obsdata/CHINA_SON2020_MP_NRT'

    reader = ReadMarcoPolo(data_dir=path_data)

    #files = reader.get_file_list()
    #print(files)

    data = reader.read(['vmro3'], last_file=1)

# =============================================================================
#     stats = data.to_station_data_all('concpm10')['stats']
#
#     stat = stats[0]
#     stats[0].plot_timeseries('concpm10')
# =============================================================================


    #data = read_cams84_china(files, ['concpm10','concpm25'])
